package edu.duke.cs.smlr;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;

import edu.duke.cs.smlr.learn.properties.KFoldProperty;
import edu.duke.cs.smlr.learn.properties.LearnProperty;
import edu.duke.cs.smlr.learn.properties.ModifyProperty;
import edu.duke.cs.smlr.learn.properties.TestProperty;
import edu.duke.cs.smlr.util.ArgumentFormatException;
import edu.duke.cs.smlr.util.Core;
import edu.duke.cs.smlr.util.Messages;
import edu.duke.cs.smlr.util.OutputController;

//TODO: javadoc this

/**Main class for the text (command line) interface.
 * 
 * @author Matt Edwards, Jason Bosko
 */

public class Text {

	/**
	 * Shows the help text whenever 'java -jar smlr.jar -help' is executed
	 * @param out The PrintStream to print the help message to.
	 */
	private static void showHelp(PrintStream out) {
		out.println("Usage: append parameters to the command.  Use the Help in the Graphical Version for more details.\n\n"
						+ "The system will use default values for parameters that are not entered.\n"
						+ "If a settings file is used, command line choices will override those in the settings file.\n\n"
						+ "When entering files, surround file names in quotes.\n\n"
						+ "Inputting files:\n"
						+ " -training FILENAME\n"
						+ " -testing FILENAME\n"
						+ " -unlabeled FILENAME\n"
						+ " -settings FILENAME\n"
						+ 
				
						"Output file: \n"
						+ " -saveto FILENAME\n"
						+

						"K-Fold settings:\n"
						+ " -kfold INTEGER\n"
						+ " -kfoldbehavior [random | ratiorandom | inplace]\n"
						+

						"Test choice setting:\n"
						+ " -behavior [testontesting | kfoldtest | loocv | unlabeled]\n"
						+

						"Setting constants:\n"
						+ " -lambda NUMBER\n"
						+ " -convergencetol NUMBER\n"
						+ " -killthreshold NUMBER\n"
						+ " -maxiterations INTEGER\n"
						+
						
						"Setting performance options:\n"
						+ " -overrelaxed [true | false]\n"
						+ " -cacheexp [true | false]\n"
						+ " -cachekernel [true | false]\n"
						+

						"Setting options:\n"
						+ " -kernel [rbf | direct | linear | cosine | polynomial]\n"
						+ " -kernelparam NUMBER\n"
						+ " -prior [laplacian | gaussian | none]\n"
						+ " -algorithm [noncomponent | component]\n"
						+ " -normalize [true | false]\n"
						+ "Setting switches:\n"
						+ " -bias [true | false]\n"
						+ "Editing data:\n"
						+ " -deleterowstrain 1 3 7 [ ... ]\n"
						+ " -deletecolumnstrain 2 4 8 [ ... ]\n"
						+ " -deleterowstest 1 3 7 [ ... ]\n"
						+ " -deletecolumnstest 2 4 8 [ ... ]\n"
						+ " -deleterowsunlabeled 1 3 7 [ ... ]\n"
						+ " -deletecolumnsunlabeled 2 4 8 [ ... ]\n"
						+ "Choosing Classifications for Training/Testing data:\n"
						+ " -choosecolumn [first | last | guess | NUMBER (one-based)]\n"
						+

						"Setting display options:\n"
						+ " -displayweights [true | false]\n"
						+ " -displayiterations [true | false]\n"
						+ " -displaybasisused [true | false]\n"
						+ " -displaytime [true | false]\n"
						+ " -displayprobabilities [true | false]\n"
						+ " -displayclassifications [true | false]\n"
						+ " -displaynumwrong [true | false]\n"
						+ " -displaynumcorrect [true | false]\n"
						+ " -displaypercentage [true | false]\n"	
						+ " -displaymisclassifications [true | false]\n"
						+
						
						"Additional display options for K-Fold or LOOCV:\n"
						+ " -displaytotalcorrect [true | false]\n"
						+ " -displaytotalwrong [true | false]\n"
						+ " -displaytotalpercentage [true | false]\n"
						+ " -displayaverageweights [true | false]\n"
						+ " -displayweightcounts [true | false]\n"
						+ " -displayaverageweightcounts [true | false]\n"
						+ " -displayconfusionmatrix [true | false]\n"
						+ " -displayroccurve [true | false]\n" +

						"Print this help message:\n" + "	-help\n" +

						"Print version information:\n" + "	-version\n");

	}

	/**
	 * Runs SMLR using the settings specified in a settings file, <b>filePath</b>, while also retaining the options specified 
	 * at the command line, contained in <b>mainOptionsSet</b> and <b>mainOptions</b>.
	 * @param filePath The settings file
	 * @param mainOptionsSet The HashSet of options specified at the command line (used for fast look-up of existance)
	 * @param mainOptions The ArrayList of options specified at the command line (used for sequential look-up)
	 */
	private static void doSettings(String filePath, HashSet mainOptionsSet,
			ArrayList mainOptions) {
		ArrayList al = new ArrayList();
		try {
		    BufferedReader input = new BufferedReader(new FileReader(filePath));		    
			String s = "";
			while (s != null) {
				s = input.readLine();
				if (s != null)
					al.add(s);
			}
		} catch (FileNotFoundException e) {
			throw new ArgumentFormatException(Messages.ERROR_FILE_NOT_FOUND);
		} catch (IOException e) {
			throw new ArgumentFormatException(Messages.ERROR_READING_FILE);
		}
		int count = al.size();
		String[] sArray = new String[count];
		for (int i = 0; i < count; i++)
			sArray[i] = (String) al.get(i);
		Core.doSettings(sArray, mainOptionsSet, mainOptions);
	}

	public static void main(String[] args) {
		ArrayList options = new ArrayList(Arrays.asList(args));
		HashSet optionsSet = new HashSet(options);
		options.add("-[null string]"); // sentinel

		ModifyProperty modifyprop[] = new ModifyProperty[3]; for (int i = 0; i < 3; i++) modifyprop[i] = new ModifyProperty();
		LearnProperty learnprop = new LearnProperty();
		TestProperty testprop = new TestProperty();
		KFoldProperty kfoldprop = new KFoldProperty();
		int[] testoption = new int[1];
		int[] columnChoice = new int[1];
		OutputController output = new OutputController();
		
		// If a settings file is given, use the options specified by it
		try {
			if (optionsSet.contains("-settings")) {
				doSettings(Core.parseFileName(options, options.indexOf("-settings") + 1), optionsSet, options);
			} else {
				testoption[0] = (Core.setDefaults(modifyprop, learnprop,
						testprop, kfoldprop, output, columnChoice));
				String outFile = Core.parseArgs(optionsSet, options, modifyprop, learnprop, testprop,
						kfoldprop, output, testoption, columnChoice);
				if (outFile == null) 
					throw new ArgumentFormatException(Messages.ERROR_NO_SAVETO);
				
				String command = ""; //the concatenation of all command line arguments
				for (int i = 0; i < args.length; ++i)
					command += args[i] + " ";
				Core.process(modifyprop, learnprop, testprop, kfoldprop, testoption[0], output, columnChoice[0], outFile, command, "");
			}
		} catch (ArgumentFormatException e) {
			if (!e.getMessage().equals("HELP"))
				System.out.println(e.getMessage()
						+ "\nUse -help to view parameter options\n");
			else {
				showHelp(System.out);
			}
		}
	}
}
