package edu.duke.cs.smlr.gui.frames;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.border.Border;
import javax.swing.border.TitledBorder;

import edu.duke.cs.smlr.Graphical;
import edu.duke.cs.smlr.gui.help.Help;
import edu.duke.cs.smlr.gui.help.HelpButton;
import edu.duke.cs.smlr.learn.properties.DisplayProperty;
import edu.duke.cs.smlr.util.Constants;
import edu.duke.cs.smlr.util.Defaults;

/**
 * Creates a panel for choosing output/export options
 * @author Matt Edwards, Jason Bosko
 */
public class OutputPreferencesFrame extends AbstractFrame {

	private JPanel kFoldPanel, testPanel, learnPanel;
	private JButton learnDefaults, testDefaults, crossDefaults;
	
	private JCheckBox learnResultDisplay[], testResultDisplay[], kFoldResultDisplay[];
	private String learnResultNames[] = { "Weights", "Iterations",
			"Number of Bases", "Time Elapsed" }, 
			testResultNames[] = {"Probabilities", "Classifications", "Number Wrong",
			"Number Correct", "Percentage", "Misclassifications" },
			kFoldResultNames[] = {"Total Correct", "Total Incorrect", 
			"Total Percentage", "Average Weights", "Weight Counts", "Nonzero Percentage",
			"Confusion Matrix", "ROC Curve"};

	private final boolean learnResultDefaults [] = {Defaults.PREFS_LEARN_WEIGHTS,
													Defaults.PREFS_LEARN_ITERS,
													Defaults.PREFS_LEARN_BASES,
													Defaults.PREFS_LEARN_TIME};
	private final boolean testResultDefaults [] = {Defaults.PREFS_TEST_PROBS,
												   Defaults.PREFS_TEST_CLASSES,
												   Defaults.PREFS_TEST_WRONG,
												   Defaults.PREFS_TEST_CORRECT,
												   Defaults.PREFS_TEST_PERCENT,
												   Defaults.PREFS_TEST_MISCLASS};
	private final boolean kFoldResultDefaults [] = {Defaults.PREFS_CROSS_CORRECT,
													Defaults.PREFS_CROSS_INCORRECT,
													Defaults.PREFS_CROSS_PERCENT,
													Defaults.PREFS_CROSS_AVGWEIGHT,
													Defaults.PREFS_CROSS_WEIGHTCOUNT,
													Defaults.PREFS_CROSS_NONZERO,
													Defaults.PREFS_CROSS_CONFUSION, 
													Defaults.PREFS_CROSS_ROC};
	private final int learnResults = 4, testResults = 6, kFoldResults = 8;
	
	protected OutputPreferencesFrame(int useless, Graphical p) {
		parent = p;
	}

	/**
	 * Constructs an OutputPreferencesFrame for the GUI.
	 * @param parent The GUI
	 */
	public OutputPreferencesFrame(Graphical parent) {
		super(parent);
		frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		frame.setSize(600, 400);
		frame.setTitle("Output Preferences");
		
		this.center();
		
		frame.getContentPane().setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		
		learnPanel = new JPanel();
		testPanel = new JPanel();
		kFoldPanel = new JPanel();

		c.gridx = 0;
		c.weightx = 1;
		c.gridy = 0;
		c.weighty = 1;
		frame.getContentPane().add(createLearnResultDisplayPanel(), c);

		c.gridy = 1;
		frame.getContentPane().add(createTestResultDisplayPanel(), c);
		
		c.gridy = 2;
		frame.getContentPane().add(createKFoldResultsPanel(), c);
		
	}

	/**
	 * Creates the panel with the options for displaying learning results
	 * @return The panel with the options for displaying learning results
	 */
	protected JPanel createLearnResultDisplayPanel() {
		learnPanel = new JPanel(new BorderLayout());
		Border blackline = BorderFactory.createLineBorder(Color.black);
		TitledBorder title = BorderFactory.createTitledBorder(blackline,
				"Output: Learning Results");
		title.setTitleJustification(TitledBorder.CENTER);
		learnPanel.setBorder(title);

		JPanel centerPanel = new JPanel(new GridLayout(2,0));
		learnResultDisplay = new JCheckBox[learnResults];
		for (int i = 0; i < learnResults; ++i) {
			learnResultDisplay[i] = new JCheckBox(learnResultNames[i], learnResultDefaults[i]);
			centerPanel.add(learnResultDisplay[i]);
		}
		
		learnDefaults = new JButton("Restore Defaults");
		//learnDefaults.setBorder(BorderFactory.createRaisedBevelBorder());
		learnDefaults.setAction(new AbstractAction("Restore Defaults") {
			public void actionPerformed(ActionEvent e) {
				for (int i = 0; i < learnResultDisplay.length; ++i)
					learnResultDisplay[i].setSelected(learnResultDefaults[i]);
			}
		});
		
		JPanel eastPanel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(13, 5, 13, 5);
		eastPanel.add(new HelpButton(Help.LEARN_DISPLAY), c);
		c.gridy = 1;
		c.weightx = 1;
		eastPanel.add(learnDefaults, c);
		
		learnPanel.add(centerPanel, BorderLayout.CENTER);
		learnPanel.add(eastPanel, BorderLayout.EAST);
		return learnPanel;
	}

	/**
	 * Creates the panel with the options for displaying testing results
	 * @return The panel with the options for displaying testing results
	 */
	protected JPanel createTestResultDisplayPanel() {
		testPanel = new JPanel(new BorderLayout());
		Border blackline = BorderFactory.createLineBorder(Color.black);
		TitledBorder title = BorderFactory.createTitledBorder(blackline,
				"Output: Testing Results");
		title.setTitleJustification(TitledBorder.CENTER);
		testPanel.setBorder(title);
		
		JPanel centerPanel = new JPanel(new GridLayout(2,0));
		testResultDisplay = new JCheckBox[testResults];
		for (int i = 0; i < testResults; ++i) {
			testResultDisplay[i] = new JCheckBox(testResultNames[i], testResultDefaults[i]);
			centerPanel.add(testResultDisplay[i]);
		}
		
		testDefaults = new JButton("Restore Defaults");
		//testDefaults.setBorder(BorderFactory.createRaisedBevelBorder());
		testDefaults.setAction(new AbstractAction("Restore Defaults") {
			public void actionPerformed(ActionEvent e) {
				for (int i = 0; i < testResultDisplay.length; ++i)
					testResultDisplay[i].setSelected(testResultDefaults[i]);				
			}			
		});
		JPanel eastPanel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(13, 5, 13, 5);
		eastPanel.add(new HelpButton(Help.TEST_DISPLAY), c);
		c.gridy = 1;
		c.weightx = 1;
		eastPanel.add(testDefaults, c);
		
		testPanel.add(centerPanel, BorderLayout.CENTER);
		testPanel.add(eastPanel, BorderLayout.EAST);
		return testPanel;
	}
	
	/**
	 * Creates the panel with the options for displaying cross-validation results
	 * @return The panel with the options for displaying cross-validation results
	 */
	protected JPanel createKFoldResultsPanel() {
		kFoldPanel = new JPanel(new BorderLayout());
		Border blackline = BorderFactory.createLineBorder(Color.black);
		TitledBorder title = BorderFactory.createTitledBorder(blackline,
				"Output: Cross Validation Results");
		title.setTitleJustification(TitledBorder.CENTER);
		kFoldPanel.setBorder(title);

		JPanel centerPanel = new JPanel(new GridLayout(3,0));
		kFoldResultDisplay = new JCheckBox[kFoldResults];
		for (int i = 0; i < kFoldResults; ++i) {
			kFoldResultDisplay[i] = new JCheckBox(kFoldResultNames[i], kFoldResultDefaults[i]);
			centerPanel.add(kFoldResultDisplay[i]);
		}
		kFoldPanel.setVisible(((Graphical) parent).getEvalChoice() != Constants.TEST_TESTING_DATA);
		
		crossDefaults = new JButton("Restore Defaults");
		//crossDefaults.setBorder(BorderFactory.createRaisedBevelBorder());
		crossDefaults.setAction(new AbstractAction("Restore Defaults") {
			public void actionPerformed(ActionEvent e) {
				for (int i = 0; i < kFoldResultDisplay.length; ++i)
					kFoldResultDisplay[i].setSelected(kFoldResultDefaults[i]);				
			}			
		});
		JPanel eastPanel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(13, 5, 13, 5);
		eastPanel.add(new HelpButton(Help.KFOLD_DISPLAY), c);
		c.gridy = 1;
		c.weightx = 1;
		eastPanel.add(crossDefaults, c);
		
		kFoldPanel.add(centerPanel, BorderLayout.CENTER);
		kFoldPanel.add(eastPanel, BorderLayout.EAST);
		return kFoldPanel;
	}

	/**
	 * Puts all the information about display options into a boolean array.
	 * The key to which display option is in which position in the array is:
	 * <br><b>Learn Results</b>
	 * <br>0: Weights
	 * <br>1: Iterations
	 * <br>2: Basis functions
	 * <br>3: Time elapsed
	 * <br><b>Single Test Results</b>
	 * <br>4: Probabilities
	 * <br>5: Classifications
	 * <br>6: Number wrong
	 * <br>7: Number correct
	 * <br>8: Percentage
	 * <br>9: Misclassifications
	 * <br><b>Summary of cross-validation</b>
	 * <br>10: Total correct
	 * <br>11: Total wrong
	 * <br>12: Total percentage
	 * <br>13: Average weights
	 * <br>14: The counts of how many times a weight was non-zero
	 * <br>15: The percentage of how many times a weight was non-zero
	 * <br>16: Confusion Matrix
	 * <br>17: ROC curve
	 * 
	 * @return The boolean array of display options.
	 */
	public boolean[] getDisplayOptions() {
		boolean opt[] = new boolean[18];
		
		opt[0] = (learnResultDisplay[0].isSelected());
		opt[1] = (learnResultDisplay[1].isSelected());
		opt[2] = (learnResultDisplay[2].isSelected());
		opt[3] = (learnResultDisplay[3].isSelected());
	
		opt[4] = (testResultDisplay[0].isSelected());
		opt[5] = (testResultDisplay[1].isSelected());
		opt[6] = (testResultDisplay[2].isSelected());
		opt[7] = (testResultDisplay[3].isSelected());
		opt[8] = (testResultDisplay[4].isSelected());
		opt[9] = (testResultDisplay[5].isSelected());
		
		opt[10] = (kFoldResultDisplay[0].isSelected());
		opt[11] = (kFoldResultDisplay[1].isSelected());
		opt[12] = (kFoldResultDisplay[2].isSelected());
		opt[13] = (kFoldResultDisplay[3].isSelected());
		opt[14] = (kFoldResultDisplay[4].isSelected());
		opt[15] = (kFoldResultDisplay[5].isSelected());
		opt[16] = (kFoldResultDisplay[6].isSelected());
		opt[17] = (kFoldResultDisplay[7].isSelected());
		
		return opt;
	}
	
	/**
	 * Sets the visibility of the k-fold display to <b>on</b>.
	 * @param on Whether the k-fold display should be visible.
	 */
	public void setKFoldDisplay(boolean on) {
		kFoldPanel.setVisible(on);
	}

	/**
	 * Sets the visibility of the test display to <b>on</b>.
	 * @param on Whether the test display should be visible.
	 */
	public void setTestDisplay(boolean on) {
		testPanel.setVisible(on);
	}

	/**
	 * Sets the display options to those specified in <b>learnDisplay</b>, <b>testDisplays</b>, and <b>crossDisplays</b>
	 * @param learnDisplays The display options for learning results
	 * @param testDisplays The display options for testing results
	 * @param crossDisplays The display options for CV results
	 */
	public void setOptions(boolean[] learnDisplays, boolean[] testDisplays, boolean[] crossDisplays) {
		for (int i = 0; i < learnDisplays.length; ++i)
			learnResultDisplay[i].setSelected(learnDisplays[i]);
		for (int i = 0; i < testDisplays.length; ++i)
			testResultDisplay[i].setSelected(testDisplays[i]);
		for (int i = 0; i < crossDisplays.length; ++i)
			kFoldResultDisplay[i].setSelected(crossDisplays[i]);
	}
}
