package edu.duke.cs.smlr.gui.frames;

import java.awt.Color;
import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.Border;
import javax.swing.border.TitledBorder;

import edu.duke.cs.smlr.Graphical;
import edu.duke.cs.smlr.gui.help.Help;
import edu.duke.cs.smlr.gui.help.HelpButton;
import edu.duke.cs.smlr.util.Defaults;

/**
 * Creates a panel for choosing techinal options in learning.
 * 
 * @author Jason Bosko, Matt Edwards
 */
public class TechnicalPreferencesFrame extends AbstractFrame {
    
    private JTextField convergenceFields[];
    private String convergenceNames[] = {"Kill Threshold","Convergence Tolerance","Maximum Iterations"};
    private final String initialConvergence[] = {Double.toString(Defaults.PREFS_KILLTHRESH),
    											 Double.toString(Defaults.PREFS_CONVTOL),
    											 Integer.toString(Defaults.PREFS_MAXITER)};
    private final int numFields = 3;
    private JCheckBox performanceBoxes[];
    private String performanceNames[] = {"Adaptive Over-Relaxation","Cache Exponentials", "Cache Kernels"};
    private boolean initialPerformance[] = {Defaults.PREFS_OVERRELAX,
    										Defaults.PREFS_CACHEEXP,
    										Defaults.PREFS_CACHEKERNEL};
    private final int numBoxes = 3;
    private JButton convergenceDefault, performanceDefault;
    
    protected TechnicalPreferencesFrame(int useless, Graphical p) {
    	parent = p;
    }
    
    /**
     * Constructs a TechnicalPreferencesFrame for the GUI.
     * @param p The GUI
     */
    public TechnicalPreferencesFrame(Graphical p) {
        super(p);
        frame.setTitle("Technical Preferences");
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        frame.setSize(500, 300);
        Container contentPane = frame.getContentPane();
        contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.Y_AXIS));
        
        this.center();
        
        contentPane.add(createConvergencePanel());
        contentPane.add(createPerformancePanel());
    }
    
    /**
     * Creates the convergence panel for the frame
     * @return The convergence panel
     */
    protected JPanel createConvergencePanel() {
        JPanel p = new JPanel(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.HORIZONTAL;
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Convergence Settings");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        
        convergenceFields = new JTextField[numFields];
        for (int i = 0; i < numFields; ++i) {
            convergenceFields[i] = new JTextField(initialConvergence[i]);
            convergenceFields[i].addFocusListener(new FocusListener() {
                public void focusGained(FocusEvent e) {
                    JTextField source = (JTextField) e.getSource();
                    source.selectAll();                    
                }
                public void focusLost(FocusEvent e) {
                    //TODO this is redundant, make this specific as we assign the listener
                    JTextField source = (JTextField) e.getSource();	
                    if (source == convergenceFields[0] || source == convergenceFields[1]) 
                        try {
                            double d = Double.parseDouble(source.getText());
                            if (d <= 0) source.setText("Cannot be negative");
                        } catch (NumberFormatException er) {
                            source.setText("Invalid Format");
                        }
                        if (source == convergenceFields[2])
                            try {
                                int ii = Integer.parseInt(source.getText());
                                if (ii <= 0) source.setText("Cannot be negative");
                            } catch (NumberFormatException er) {
                                source.setText("Invalid Format");
                            }                    
                }
                
            });
            c.gridx = 0;
            c.gridy = i;
            c.weightx = 0;
            p.add(new JLabel(convergenceNames[i]), c);
            c.gridx = 1;
            c.gridy = i;
            c.weightx = 1;
            p.add(convergenceFields[i], c);
        }
        
        convergenceDefault = new JButton("Restore Defaults");
        //convergenceDefault.setBorder(BorderFactory.createRaisedBevelBorder());
        convergenceDefault.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                for (int i = 0; i < numFields; ++i)
                    convergenceFields[i].setText(initialConvergence[i]);
            }
        });
        c.gridx = 2; c.gridy = numFields-1; c.weightx = 0;
        p.add(convergenceDefault, c);
        
        c.gridx = 3; c.gridy = 0; c.weightx = 0;
        p.add(new HelpButton(Help.CONVERGENCE_SETTINGS));
        return p;
    }
    
    
    /**
     * Creates the performance panel for the frame
     * @return The performance panel
     */
    protected JPanel createPerformancePanel() {
        JPanel p = new JPanel(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.HORIZONTAL;
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Performance Settings");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        
        performanceBoxes = new JCheckBox[numBoxes];
        for (int i = 0; i < numBoxes; ++i) {
            performanceBoxes[i] = new JCheckBox(performanceNames[i], initialPerformance[i]);
            c.gridx = 0;
            c.gridy = i;
            c.weightx = 1;
            p.add(performanceBoxes[i], c);
        }
        
        performanceDefault = new JButton("Restore Defaults");
        //performanceDefault.setBorder(BorderFactory.createRaisedBevelBorder());
        performanceDefault.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                for (int i = 0; i < numBoxes; ++i)
                    performanceBoxes[i].setSelected(initialPerformance[i]);                
            }
        });
        c.gridx = 1; c.gridy = numBoxes-1; c.weightx = 0;
        p.add(performanceDefault, c);
        
        c.gridx = 2; c.gridy = 0; c.weightx = 0;
        p.add(new HelpButton(Help.PERFORMANCE_SETTINGS));
        
        return p;
    }
    
    public void actionPerformed(ActionEvent e) {
        Object source = e.getSource();
        if (source == convergenceDefault) 
            for (int i = 0; i < numFields; ++i)
                convergenceFields[i].setText(initialConvergence[i]);
        if (source == performanceDefault)
            for (int i = 0; i < numBoxes; ++i)
                performanceBoxes[i].setSelected(initialPerformance[i]);
    }	
    
    /**
     * Gets the kill threshold specified in this frame.
     * @return the kill threshold
     */
    public double getKillThreshold() { 
    	//System.out.println("IN");
    	//System.out.println(convergenceFields[0].getText());    	
        try {
            return Double.parseDouble(convergenceFields[0].getText());
        } catch (NumberFormatException e) {        	
        	//System.out.println(e.getMessage());
            return Double.parseDouble(initialConvergence[0]);
        }
    }
    
    /**
     * Gets the convergence tolerance specified in this frame.
     * @return the convergence tolerance
     */
    public double getConvergenceTolerance() {
        try {
            return Double.parseDouble(convergenceFields[1].getText());
        } catch (NumberFormatException e) {
            return Double.parseDouble(initialConvergence[1]);
        }
    }
    /**
     * Gets the maximum iterations specified in this frame.
     * @return the maximum iterations
     */
    public int getMaximumIterations() {
        try {
            return Integer.parseInt(convergenceFields[2].getText());
        } catch (NumberFormatException e) {
            return Integer.parseInt(initialConvergence[2]);
        }
    }
    
    /**
     * Reports whether overrelaxation has been selected
     * @return whether overrelaxation has been selected
     */
    public boolean isOverrelaxation() {
        return performanceBoxes[0].isSelected();
    }
    /**
     * Reports whether caching exponentials has been selected
     * @return whether caching exponentials has been selected
     */
    public boolean isCacheExponentials() {
        return performanceBoxes[1].isSelected();
    }
    /**
     * Reports whether caching kernels has been selected
     * @return whether caching kernels has been selected
     */
    public boolean isCacheKernels() {
    	return performanceBoxes[2].isSelected();
    }
    
    /**
     * Sets the options available to those specified by the parameters
     * @param killThresh The kill threshold to be set
     * @param convTol The convergence tolerance to be set
     * @param maxIter The maximum iterations to be set
     * @param overrelax Whether to use overrelxation
     * @param cacheExp Whether to cache exponentials
     * @param cacheKer Whether to cache kernels
     */
    public void setOptions(double killThresh, double convTol, int maxIter, boolean overrelax, boolean cacheExp, boolean cacheKer) {
        convergenceFields[0].setText(""+killThresh);
        convergenceFields[1].setText(""+convTol);
        convergenceFields[2].setText(""+maxIter);
        performanceBoxes[0].setSelected(overrelax);
        performanceBoxes[1].setSelected(cacheExp);
        performanceBoxes[2].setSelected(cacheKer);
    }
}
