package edu.duke.cs.smlr.gui.panels;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.border.Border;
import javax.swing.border.TitledBorder;

import edu.duke.cs.smlr.Graphical;
import edu.duke.cs.smlr.gui.help.Help;
import edu.duke.cs.smlr.gui.help.HelpButton;
import edu.duke.cs.smlr.gui.table.AlternatingRowRenderer;
import edu.duke.cs.smlr.gui.table.BooleanRenderer;
import edu.duke.cs.smlr.gui.table.UneditableModel;
import edu.duke.cs.smlr.learn.properties.LearnProperty;
import edu.duke.cs.smlr.util.ArgumentFormatException;
import edu.duke.cs.smlr.util.Constants;
import edu.duke.cs.smlr.util.Core;
import edu.duke.cs.smlr.util.HaltingException;
import edu.duke.cs.smlr.util.Messages;

/**
 * A panel for displaying all combinations of possible tests and allowing the user to select/deselect those tests before learning.
 * @author Jason Bosko, Matt Edwards
 */

public class ChoosePanel extends AbstractPanel implements ActionListener, FocusListener {
	final private boolean GRAPHICAL_YES = false;
    private LearnProperty[] learnprops;
    
    final private String yes = "Yes";
    final private String no = "No";
    private Object[][] values;
    private HashSet unselected;
    
    private JTable table;
    private JScrollPane pane;
    
    private JButton learn;
    private JButton save;    
    
    private JTextField fileLabel;
	private JButton fileBrowser;
	private String file;
	
	private JLabel learningInProgress = new JLabel();
	private JLabel testCountLabel = new JLabel();
	private int numberTestsInProgress = 0;
    
    /**
     * Capitalizes the string <b>s</b>. Also turns "rbf" into "RBF".
     * @param s The string to be capitalized.
     * @return The capitalized string.
     */
	private static String capitalize(String s) {
        if(s == null || s.length() == 0) return s;
        if(s.equals("rbf")) return "RBF";
        return s.substring(0, 1).toUpperCase() + s.substring(1, s.length());
    }
    
    /**This updates and shows the specified learn properties in the table.
     * @param l the learnProperty array to update with
     */
    public void update(LearnProperty[] l) {
        learnprops = l;
        if (learnprops.length > 1) 
        	testCountLabel = new JLabel(Messages.NOTIFY_TESTS_CHOSEN);
        else
        	testCountLabel = new JLabel(Messages.NOTIFY_ONE_POSSIBLE_TEST);
        if(pane != null) panel.removeAll();        
        
        String[] names = {"Selected", "Lambda", "Prior", "Update Rule", "Kernel [Param]", "Normalize", "Bias Term"};
        values = new Object[learnprops.length][names.length];
        for(int i = 0; i < learnprops.length; ++i) {
            values[i][0] = new Boolean(!unselected.contains(new Integer(learnprops[i].hashCode())));
            values[i][1] = Double.toString((learnprops[i].getLambda()));
            values[i][4] = capitalize(Constants.toString(learnprops[i].getKernel()));
            if(learnprops[i].getKernel() == Constants.KERNEL_POLYNOMIAL || learnprops[i].getKernel() == Constants.KERNEL_RBF)
                values[i][4] = (String) values[i][4] + " [ " + learnprops[i].getKernelPar() + " ]";
            values[i][3] = capitalize(Constants.toString(learnprops[i].getMethod()));
            values[i][2] = capitalize(Constants.toString(learnprops[i].getPrior()));
            
            if (GRAPHICAL_YES) {
            	//values[i][6] = new Boolean(learnprops[i].isBias());
            	values[i][5] = new Boolean(learnprops[i].isNormalize());
            }
            else {
            	//values[i][6] = learnprops[i].isBias() ? yes : no;
            	values[i][5] = learnprops[i].isNormalize() ? yes : no;            	
            }	
            values[i][6] = learnprops[i].getBias() != 0 ? Double.toString(learnprops[i].getBias()) : "none";
        }
        
        table = new JTable(new UneditableModel(values, names)) {
            public boolean isCellEditable(int row, int col) {
                return col == 0;
            }                 
        };
        
        table.setDefaultRenderer(Object.class, new AlternatingRowRenderer());
        if (GRAPHICAL_YES) {
        	JRadioButton b = new JRadioButton();
            b.setEnabled(false);
        	table.getColumnModel().getColumn(5).setCellRenderer(new BooleanRenderer(b));
        	table.getColumnModel().getColumn(6).setCellRenderer(new BooleanRenderer(b));
        }
        table.getColumnModel().getColumn(0).setCellRenderer(new BooleanRenderer(new JCheckBox()));
        
        for(int i = 0; i < table.getColumnCount(); ++i)
            table.getColumnModel().getColumn(i).sizeWidthToFit();
        
        table.setPreferredScrollableViewportSize(new Dimension(400, 200));
        pane = new JScrollPane(table, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        
        
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        
        c.weightx = 1;
        panel.add(testCountLabel, c);
        
        c.gridy = 1;
        c.weighty = 1;
        panel.add(pane, c);
        
        JPanel southPanel = new JPanel(new GridBagLayout());

        GridBagConstraints c2 = new GridBagConstraints();
        c2.fill = GridBagConstraints.HORIZONTAL;
        c2.weightx = 1;
        c2.gridwidth = 2;
        southPanel.add(createSaveResultsPanel(), c2);

        c2.gridx = 2;
        c2.weightx = 0;
        c2.gridwidth = 1;
        c2.insets = new Insets(0, 5, 0, 5);
        southPanel.add(new HelpButton(Help.CHOOSE_PANEL), c2);
        
        c2.insets = new Insets(20, 20, 0, 20);
        c2.gridx = 0;
        c2.weightx = 1;
        c2.gridy = 1;
        southPanel.add(learn, c2);
        
        c2.gridx = 1;
        southPanel.add(save, c2);
        
        c2.gridx = 2;
        c2.weightx = 0;
        c2.insets = new Insets(20, 5, 0, 5);
        southPanel.add(new HelpButton(Help.SAVE_AND_LEARN), c2);
        
        c2.gridx = 0;
        c2.gridy = 2;
        c2.gridwidth = GridBagConstraints.REMAINDER;
        southPanel.add(learningInProgress, c2);
        
        updateLearningInProgress();
        
        c.gridy = 2;
        c.weighty = 0.3;
        panel.add(southPanel, c);
        
        
        panel.validate();
    }
    
    /**Updates the private pool of unselected learning tasks.  This must be called
     * when focus leaves this panel in order for the unselected tasks to be properly
     * recorded.  This must be called using the same learnPropery array that was
     * passed to update, or else the wrong tasks will be recorded as selected or
     * unselected. 
     */
    public void updateUnselected() {
        unselected.clear();
        for(int i = 0; i < learnprops.length; ++i)
            if (!((Boolean) values[i][0]).booleanValue())
                unselected.add(new Integer(learnprops[i].hashCode()));            
    }
        
    /**Assembles the selected learning tasks from the panel and returns them as
     * an array of strings, each representing the proper parameters to pass to
     * Core.  This method gets information from Graphical to fill in the appropriate
     * fields.
     * @return an array of learning task parameters
     */
    public String[] tasks() {
        ArrayList ans = new ArrayList();
        for(int i = 0; i < values.length; ++i)
            if(((Boolean) values[i][0]).booleanValue())
                ans.add(parent.getSettings(learnprops[i]));
        return (String[]) ans.toArray(new String[0]);
    }
    
    /**
     * Creates the panel with the text field for choosing a file to save results to.
     * @return the panel with the save results text field.
     */
    private JPanel createSaveResultsPanel() {
		JPanel p = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		Border blackline = BorderFactory.createLineBorder(Color.black);
		TitledBorder title = BorderFactory.createTitledBorder(blackline, "Save Results");
		title.setTitleJustification(TitledBorder.CENTER);
		p.setBorder(title);
		
		fileLabel = new JTextField(file);
		fileLabel.addFocusListener(this);
		fileBrowser = new JButton(" Browse ");
		//fileBrowser.setBorder(BorderFactory.createRaisedBevelBorder());
		fileBrowser.addActionListener(this);
		
		c.insets = new Insets(0, 5, 0, 5);
		
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 0;
		p.add(new JLabel("Directory to Save Results to: "), c);
		
		c.gridx = 2;
		c.weightx = 1;
		p.add(fileLabel, c);
		
		c.gridx = 4;
		c.weightx = 0;
		p.add(fileBrowser, c);
		return p;
	}
    
    /**
     * Constructs the choose panel for the GUI.
     * @param p The GUI.
     */
    public ChoosePanel(Graphical p) {
        super(p);
        
        unselected = new HashSet();
        file = "";
        panel.setLayout(new GridBagLayout());
        
        learn = new JButton("Learn Classifier");
        //learn.setBorder(BorderFactory.createRaisedBevelBorder());
        learn.addActionListener(this);
        
        save = new JButton("Save Settings");
        //save.setBorder(BorderFactory.createRaisedBevelBorder());
        save.addActionListener(this);
        
        panel.setEnabled(true);
        panel.setVisible(true);
        
        this.update(new LearnProperty[0]);
    }
    
    public void actionPerformed(ActionEvent evt) {
        Object o = evt.getSource();
        
        if (o == learn) {
        	if (!validateResultsDir(file)) {
    			JOptionPane.showMessageDialog(panel, "Error: You must choose a directory to save the results to.", "Error", JOptionPane.ERROR_MESSAGE);
    			return;
    		}
            if (((Graphical) parent).isTrainingLoaded()) {
				if (!fileLabel.getText().equals(Constants.NO_FILE_LOADED)) {	
            		try {
						Core.doSettings(tasks(), parent);
						JOptionPane.showMessageDialog(parent,
								"The learning task was started.  \n" +
								"If you close the program before it completes, the task will be stopped. \n" +
								"If you start another task using the same output file as this one before it completes, \n" +
								"the file will be unreadable.  A message will pop up when this task is finished.",
								"Learning Started", JOptionPane.INFORMATION_MESSAGE);
						numberTestsInProgress++;
						updateLearningInProgress();
						learningInProgress.setVisible(true);
						//learn.setEnabled(false);

					}
					catch (ArgumentFormatException e) {
						JOptionPane.showMessageDialog(panel,
								"Error: you did not provide enough information to begin the learning task.",
								"Cannot Start", JOptionPane.ERROR_MESSAGE);
					}
					catch (HaltingException e) {
						JOptionPane.showMessageDialog(panel,
								"Error encountered during learning task. Please restart.",
								"Cannot Start", JOptionPane.ERROR_MESSAGE);
					}
				}
				else 
					JOptionPane.showMessageDialog(panel, 
            				"Error: No output file has been selected. \n" + 
            				"Please choose a file where it says \"Save Results To\"",
							"Error", JOptionPane.ERROR_MESSAGE);
            }
            else 
            	JOptionPane.showMessageDialog(panel, 
            				"Error: No training data was loaded. \n" + 
            				"Please return to the \"Specify Task\" tab.",
							"Error", JOptionPane.ERROR_MESSAGE);
        }
        if (o == save) {
        	file = fileLabel.getText();
        	JFileChooser fc = edu.duke.cs.smlr.util.Utilities.newFileChooser();
			if (fc.showSaveDialog(panel) == JFileChooser.APPROVE_OPTION) {
				((Graphical) parent).saveSettings(fc.getSelectedFile().getPath(), false);
			}
			edu.duke.cs.smlr.util.Utilities.setDirectory(fc.getCurrentDirectory());			
        }
        
        if (o == fileBrowser) {
        	JFileChooser fc = edu.duke.cs.smlr.util.Utilities.newFileChooser();
        	fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			if (fc.showSaveDialog(panel) == JFileChooser.APPROVE_OPTION) {
				file = fc.getSelectedFile().getAbsolutePath();
				fileLabel.setText(file);
			}
			edu.duke.cs.smlr.util.Utilities.setDirectory(fc.getCurrentDirectory());			
        } 
    }
    
    final static String [] numeros = {"No", "One", "Two", "Three", "Four", "Five", "Six", "Seven", "Eight", "Nine", "Ten"};

	/**
	 * Updates the string informing the user of how many tests are in progress.
	 *
	 */
    private void updateLearningInProgress() {
		String text;
		if (numberTestsInProgress >= 0 && numberTestsInProgress < numeros.length)
			text = numeros[numberTestsInProgress];
		else
			text = Integer.toString(numberTestsInProgress);
		if (numberTestsInProgress == 1)
			text += " learning task currently in progress...";
		else
			text += " learning tasks currently in progress...";
		learningInProgress.setText(text);
		
		if (numberTestsInProgress <= 0)
			learningInProgress.setVisible(false);
	}

	public void focusGained(FocusEvent e) {
		((JTextField) e.getSource()).selectAll();
	}

	public void focusLost(FocusEvent e) {
		file = ((JTextField) e.getSource()).getText();
	}
	
	/**
	 * Checks to see if <b>s</b> is a valid directory (either does not exist yet or is a current directory).
	 * @param s The "directory" in question.
	 * @return Whether <b>s</b> is a valid directory name.
	 */
	private boolean validateResultsDir(String s) {
		File f = new File(s);
		return !f.exists() || f.isDirectory();
	}
	/**Gets the the filename of the path where the user wishes to save results
	 * @return the filename
	 */
	public String getSaveToFile() {
		//return file;
		return fileLabel.getText();
	}
	
	/**Sets the filename of the path where the user wishes to save results
	 * @param saveToFile the filename
	 */
	public void setOptions(String saveToFile) {
		file = saveToFile;
		fileLabel.setText(saveToFile);
	}	

	/**
	 * The name is misleading -- simply informs the panel that a test has finished.
	 *
	 */
	public void reenableLearn() {
		numberTestsInProgress--;
		updateLearningInProgress();
	}
}
