package edu.duke.cs.smlr.gui.panels;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.TreeSet;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.Border;
import javax.swing.border.TitledBorder;

import edu.duke.cs.smlr.Graphical;
import edu.duke.cs.smlr.gui.help.Help;
import edu.duke.cs.smlr.gui.help.HelpButton;
import edu.duke.cs.smlr.util.Constants;
import edu.duke.cs.smlr.util.Defaults;
import edu.duke.cs.smlr.util.Messages;

/**Creates a panel for setting the methods (algorithm, kernel, and associated options)
 * 
 * @author Matt Edwards, Jason Bosko 
 */

public class MethodsPanel extends AbstractPanel implements ItemListener, FocusListener {
    
    private JCheckBox cbPriors[];
    private JCheckBox cbStep, cbIter;
    private JCheckBox cbBiasYes, cbBiasNo;
    private JCheckBox cbKernels[];
    private JCheckBox cbNormalizeYes;
    private JCheckBox cbNormalizeNo;
    
    private JTextField polyParamField;
    private JTextField rbfParamField;
    private JTextField lambdaField;
    
    private JTextField biasField;
    
    private JLabel polyLabel;
    private JLabel rbfLabel;
    
    final private String priorNames[] = { "Laplacian", "Gaussian", "None" };
    final private String kernelNames[] = { "Direct", "Linear", "Polynomial", "RBF", "Cosine" };
    
    final private boolean priorDefaults[] = {Defaults.PRIOR == Constants.PRIOR_LAPLACIAN,
    										 Defaults.PRIOR == Constants.PRIOR_GAUSSIAN,
    										 Defaults.PRIOR == Constants.PRIOR_NONE};
    final private boolean kernelDefaults[] = {Defaults.KERNEL == Constants.KERNEL_DIRECT,
    										  Defaults.KERNEL == Constants.KERNEL_LINEAR,
    										  Defaults.KERNEL == Constants.KERNEL_POLYNOMIAL,
    										  Defaults.KERNEL == Constants.KERNEL_RBF,
    										  Defaults.KERNEL == Constants.KERNEL_COSINE};
    					
    /**
     * Constructs a methods panel for the GUI.
     * @param g The GUI.
     */
    public MethodsPanel(Graphical g) {
        super(g);
        panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
        panel.setBackground(Color.gray);
        
        JPanel leftPanel = new JPanel();
        leftPanel.setLayout(new BoxLayout(leftPanel, BoxLayout.Y_AXIS));
        JPanel rightPanel = new JPanel();
        rightPanel.setLayout(new BoxLayout(rightPanel, BoxLayout.Y_AXIS));
        
        leftPanel.add(createNormalizePanel());
        leftPanel.add(createKernel());
        
        rightPanel.add(createBias());
        rightPanel.add(createPrior());
        rightPanel.add(createLambdaPanel());
        rightPanel.add(createAlgorithm());
        
        panel.add(leftPanel);
        panel.add(rightPanel);
    }
    
    /**
     * Creates the panel for prior options
     * @return The panel for prior options.
     */
    private JPanel createPrior() {
        JPanel p = new JPanel(new BorderLayout());
        p.setName("Prior");
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Prior");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        
        int priorNum = priorNames.length;
        cbPriors = new JCheckBox[priorNum];
        
        JPanel centerPanel = new JPanel(new GridLayout(3,1));
        for (int k = 0; k < priorNum; ++k) {
            cbPriors[k] = new JCheckBox(priorNames[k], priorDefaults[k]);
            cbPriors[k].addItemListener(this);
            centerPanel.add(cbPriors[k]);
        }
        JPanel eastPanel = new JPanel();
        eastPanel.add(new HelpButton(Help.PRIORS));
        p.add(eastPanel, BorderLayout.EAST);
        p.add(centerPanel, BorderLayout.CENTER);
        return p;
    }
    
    /**
     * Creates the panel for algorithm options
     * @return The panel for algorithm options.
     */
    private JPanel createAlgorithm() {
        JPanel p = new JPanel(new BorderLayout());
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Update Rule");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        
        cbStep = new JCheckBox("Component-Wise", Defaults.ALGORITHM == Constants.ALGORITHM_COMPONENT);
        cbIter = new JCheckBox("Non-Component-Wise", Defaults.ALGORITHM == Constants.ALGORITHM_NON_COMPONENT);
        
        JPanel centerPanel = new JPanel(new GridLayout(2, 1));
        centerPanel.add(cbStep);
        centerPanel.add(cbIter);
        JPanel eastPanel = new JPanel();
        eastPanel.add(new HelpButton(Help.UPDATE_RULE));
        
        p.add(centerPanel, BorderLayout.CENTER);
        p.add(eastPanel, BorderLayout.EAST);
        return p;
    }
    
    /**
     * Creates the panel for kernel options
     * @return The panel for kernel options.
     */
    private JPanel createKernel() {
        JPanel p = new JPanel(new BorderLayout());
        p.setName("Kernels");
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Kernel Choice");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        
        JPanel kerns = new JPanel(new GridLayout(5, 1));
        int numKernels = kernelNames.length;
        cbKernels = new JCheckBox[numKernels];
        
        for (int k = 0; k < numKernels; k++) {
            cbKernels[k] = new JCheckBox(kernelNames[k], kernelDefaults[k]);
            cbKernels[k].addItemListener(this);
            kerns.add(cbKernels[k]);
        }
        
        JPanel params = new JPanel(new GridLayout(5, 1));
        
        polyParamField = new JTextField(Double.toString(Defaults.KERNEL_POLY_PARAM));
        polyParamField.setVisible(Defaults.KERNEL == Constants.KERNEL_POLYNOMIAL);
        polyParamField.addFocusListener(this);
        
        rbfParamField = new JTextField(Double.toString(Defaults.KERNEL_RBF_PARAM));
        rbfParamField.setVisible(Defaults.KERNEL == Constants.KERNEL_RBF);
        rbfParamField.addFocusListener(this);
        
        polyLabel = new JLabel("Enter Degree(s) Below");
        polyLabel.setVisible(false);
        JPanel pPanel = new JPanel(new GridLayout(2, 1));
        pPanel.add(polyLabel);
        pPanel.add(polyParamField);
        
        rbfLabel = new JLabel("Enter Bandwidth(s) Below");
        rbfLabel.setVisible(false);
        JPanel rPanel = new JPanel(new GridLayout(2, 1));
        rPanel.add(rbfLabel);
        rPanel.add(rbfParamField);
        
        params.add(new JPanel());
        params.add(new JPanel());
        params.add(pPanel);
        params.add(rPanel);
        params.add(new JPanel());
        
        JPanel centerPanel = new JPanel(new BorderLayout(20, 0));
        centerPanel.add(kerns, BorderLayout.WEST);
        centerPanel.add(params, BorderLayout.CENTER);
        JPanel eastPanel = new JPanel();
        eastPanel.add(new HelpButton(Help.KERNELS));
        
        p.add(centerPanel, BorderLayout.CENTER);
        p.add(eastPanel, BorderLayout.EAST);
        return p; 
    }
    
    /**
     * Creates the panel for bias options
     * @return The panel for bias options.
     */
    private JPanel createBias() {
        JPanel p = new JPanel(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        p.setName("Bias");
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Bias Term");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        
        cbBiasYes = new JCheckBox("Add Bias(es) ", Defaults.BIAS != 0);
        cbBiasNo = new JCheckBox("Do Not Add Bias", Defaults.BIAS == 0);
        
        biasField = new JTextField(Defaults.BIAS != 0 ? Double.toString(Defaults.BIAS) : "", 10);
        biasField.setVisible(Defaults.BIAS != 0);
        biasField.addFocusListener(this);
        
        cbBiasYes.addItemListener(this);
        
        c.fill = GridBagConstraints.HORIZONTAL;
        c.gridx = 0;
        c.gridy = 0;
        c.weightx = 0;
        p.add(cbBiasYes, c);
        c.gridx = 1;
        c.weightx = 1;
        p.add(biasField, c);
        
        c.gridy = 1;
        c.gridx = 0;
        c.gridwidth = 2;
        p.add(cbBiasNo, c);
        
        c.insets = new Insets(5,5,5,5);
        c.gridx = 2;
        c.gridy = 0;
        c.weightx = 0;
        c.gridwidth = 0;
        p.add(new HelpButton(Help.BIAS), c);
        
        return p;
    }
    
    /**
     * Creates the panel for normalize options
     * @return The panel for normalize options.
     */
    private JPanel createNormalizePanel() {
        JPanel p = new JPanel(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Data Normalization");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        
        cbNormalizeYes = new JCheckBox("Normalize", Defaults.NORMALIZE);
        cbNormalizeNo = new JCheckBox("Do Not Normalize", !Defaults.NORMALIZE);
        
        c.fill = GridBagConstraints.HORIZONTAL;
        c.gridx = 0;
        c.gridy = 0;
        c.weightx = 1;
        p.add(cbNormalizeYes, c);
        c.gridy = 1;
        p.add(cbNormalizeNo, c);
        
        c.insets = new Insets(5,5,5,5);
        c.gridx = 1;
        c.gridy = 0;
        c.weightx = 0;
        p.add(new HelpButton(Help.NORMALIZE), c);
        
        return p;
    }
    
    /**
     * Creates the panel for lambda.
     * @return The panel for lambda.
     */
    private JPanel createLambdaPanel() {
        JPanel p = new JPanel(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        Border blackline = BorderFactory.createLineBorder(Color.black);
        TitledBorder title = BorderFactory.createTitledBorder(blackline, "Sparsity Control");
        title.setTitleJustification(TitledBorder.CENTER);
        p.setBorder(title);
        lambdaField = new JTextField(10);
        lambdaField.addFocusListener(this);
        
        c.fill = GridBagConstraints.HORIZONTAL;
        c.weightx = 0;
        c.weighty = 0;
        c.gridx = 0;
        c.gridy = 0;
        p.add(new JLabel("Lambda(s): "), c);
        
        c.weightx = 1;
        c.gridx = 1;
        p.add(lambdaField, c);
        
        c.insets = new Insets(5,5,5,5);
        c.weightx = 0;
        c.gridx = 2;
        p.add(new HelpButton(Help.SPARSITY), c);
        return p;
    }
    
    public void itemStateChanged(ItemEvent e) {
        Object source = e.getItemSelectable();
        if (source == cbKernels[2]) {
            polyParamField.setVisible(!polyParamField.isVisible());
            polyLabel.setVisible(polyParamField.isVisible());
        }
        if (source == cbKernels[3]) {
            rbfParamField.setVisible(!rbfParamField.isVisible());
            rbfLabel.setVisible(rbfParamField.isVisible());
        }
        if (source == cbBiasYes) {
        	biasField.setVisible(cbBiasYes.isSelected());
        }
    }
    
    /**
     * Gets the priors that have been selected and puts them in an array.
     * @return The array of priors that are selected.
     */
    public int[] getPriors() {
        int consts[] = { Constants.PRIOR_LAPLACIAN,
                Constants.PRIOR_GAUSSIAN, Constants.PRIOR_NONE };
        int checked = 0;
        int prs[] = new int[3];
        
        for (int k = 0; k < 3; ++k)
            if (cbPriors[k].isSelected()) {
                prs[checked] = consts[k];
                checked++;
            }
            
        if (checked == 0) {
            checked++;
            int defIndex = 0;
            if (Defaults.PRIOR == Constants.PRIOR_LAPLACIAN) defIndex = 0;
            if (Defaults.PRIOR == Constants.PRIOR_GAUSSIAN) defIndex = 0;
            if (Defaults.PRIOR == Constants.PRIOR_NONE) defIndex = 0;
            cbPriors[defIndex].setSelected(true);
            prs[0] = consts[defIndex];
        }
        
        int priors[] = new int[checked];
        for (int k = 0; k < checked; ++k)
            priors[k] = prs[k];
        
        return priors;
    }
    
    /**Returns an integer array of the selected kernels.
     * @return an int[] of the selected kernels
     */
    public int[] getKernels() {
        int consts[] = { Constants.KERNEL_DIRECT,
                Constants.KERNEL_LINEAR,
                Constants.KERNEL_POLYNOMIAL,
                Constants.KERNEL_RBF,
                Constants.KERNEL_COSINE};
        int checked = 0;
        int kers[] = new int[5];
        
        for (int k = 0; k < 5; ++k)
            if (cbKernels[k].isSelected()) {
                kers[checked] = consts[k];
                checked++;
            }
        int defIndex = 0;
        if (Defaults.KERNEL == Constants.KERNEL_DIRECT) defIndex = 0;
        if (Defaults.KERNEL == Constants.KERNEL_LINEAR) defIndex = 1;
        if (Defaults.KERNEL == Constants.KERNEL_POLYNOMIAL) defIndex = 2;
        if (Defaults.KERNEL == Constants.KERNEL_RBF) defIndex = 3;
        if (Defaults.KERNEL == Constants.KERNEL_COSINE) defIndex = 4;
        if (checked == 0) {
            checked++;
            cbKernels[defIndex].setSelected(true);
            kers[0] = consts[defIndex];
        }
        
        int kernels[] = new int[checked];
        for (int k = 0; k < checked; ++k)
            kernels[k] = kers[k];
        
        return kernels;
    }
    
    /**Returns an integer array of the selected algorithms
     * @return an int[] of the selected algorithms
     */
    public int[] getAlgorithms() {
        int checked = 0;
        if (cbIter.isSelected())
            checked++;
        if (cbStep.isSelected())
            checked++;
        
        if (checked == 0) {
            checked++;
            if (Defaults.ALGORITHM == Constants.ALGORITHM_COMPONENT)
            	cbStep.setSelected(true);
            else cbIter.setSelected(true);
        }
        
        int algorithms[] = new int[checked];
        int index = 0;
        if (cbIter.isSelected()) {
            algorithms[index] = Constants.ALGORITHM_NON_COMPONENT;
            index++;
        }
        if (cbStep.isSelected()) {
            algorithms[index] = Constants.ALGORITHM_COMPONENT;
            index++;
        }
        
        return algorithms;
    }
    
    /**
     * Gets the array of bias terms to use. A bias term of zero indicates that no bias should be used.
     * @return The array of bias terms.
     */
    public double[] getBias() {
    	try {
    		if (!cbBiasNo.isSelected() && !cbBiasYes.isSelected()) {
    			double d[] = new double[1];
    			d[0] = Defaults.BIAS;
    			return d;
    		}
    		int count = 0;
    		if (cbBiasNo.isSelected()) {
    			count++;
    		}	
    		if (cbBiasYes.isSelected()) {
    			String[] args = biasField.getText().split("[, ]");
    			for (int i = 0; i < args.length; ++i)
    				if (!args[i].equals("")) 
    					count++;
    		}
    		if (count == 0) {
    			double d[] = new double[1];
    			d[0] = Defaults.BIAS;
    			return d;
    		}
    		
    		int index = 0;
    		double d[] = new double[count];
    		if (cbBiasNo.isSelected()) {
    			d[index] = 0;
    			index++;
    		}
    		if (cbBiasYes.isSelected()) {
    			String[] args = biasField.getText().split("[, ]");
    			for (int i = 0; i < args.length; ++i) 
    				if (!args[i].equals("")) {
    					d[index] = Double.parseDouble(args[i]);
    					index++;
    				}
    		}
    		return d;
    	} catch (NumberFormatException ex) {
    		polyParamField.setText(Messages.ERROR_COMMA_LIST);
    		double d[] = new double[1];
    		d[0] = Defaults.KERNEL_POLY_PARAM;
    		return d;
    	}
    }
    
    /**
     * @return whether "normalize" is set
     */
    public boolean getNormalizeYes() {
        return cbNormalizeYes.isSelected();
    }
    
    /**
     * @return whether "do not normalize" is set
     */
    public boolean getNormalizeNo() {
        return cbNormalizeNo.isSelected();
    }
    
    /**Returns an integer array of the polynomial kernel exponents
     * @return an int[] of the kernel exponents
     */
    public double[] getPolyPars() {
        try {
            String[] args = polyParamField.getText().split("[, ]");
            int count = 0;
            for (int i = 0; i < args.length; ++i)
                if (!args[i].equals("")) 
                    count++;
            if (count == 0) {
                double d[] = new double[1];
                d[0] = Defaults.KERNEL_POLY_PARAM;
                return d;
            }
            int index = 0;
            double d[] = new double[count];
            for (int i = 0; i < args.length; ++i) 
                if (!args[i].equals("")) {
                    d[index] = Double.parseDouble(args[i]);
                    index++;
                }
            return d;
        } catch (NumberFormatException ex) {
            polyParamField.setText("Use a comma-separated list");
            double d[] = new double[1];
            d[0] = Defaults.KERNEL_POLY_PARAM;
            return d;
        }
    }
    
    /**Returns an integer array of the RBF kernel parameters
     * @return an int[] of the RBF bandwidths
     */
    public double[] getRBFPars() {
        try {
            String[] args = rbfParamField.getText().split("[, ]");
            int count = 0;
            for (int i = 0; i < args.length; ++i)
                if (!args[i].equals("")) 
                    count++;
            int index = 0;
            if (count == 0) {
                double d[] = new double[1];
                d[0] = Defaults.KERNEL_RBF_PARAM;
                return d;
            }
            double d[] = new double[count];
            for (int i = 0; i < args.length; ++i) 
                if (!args[i].equals("")) {
                    d[index] = Double.parseDouble(args[i]);
                    index++;
                }
            return d;
        } catch (NumberFormatException ex) {
            rbfParamField.setText(Messages.ERROR_COMMA_LIST);
            double d[] = new double[1];
            d[0] = Defaults.KERNEL_RBF_PARAM;
            return d;
        }
    }
    
    /**Returns an integer array of the selected lambdas
     * @return an int[] of the lambdas
     */
    public double[] getLambdas() {
        try {
            String[] args = lambdaField.getText().split("[, ]");
            int count = 0;
            for (int i = 0; i < args.length; ++i)
                if (!args[i].equals("")) 
                    count++;
            int index = 0;
            if (count == 0) {
                double d[] = new double[1];
                d[0] = Defaults.LAMBDA;
                return d;
            }
            double d[] = new double[count];
            for (int i = 0; i < args.length; ++i) 
                if (!args[i].equals("")) {
                    d[index] = Double.parseDouble(args[i]);
                    index++;
                }
            return d;
        } catch (NumberFormatException ex) {
            lambdaField.setText(Messages.ERROR_COMMA_LIST);
            double d[] = new double[1];
            d[0] = Defaults.LAMBDA;
            return d;
        }
    }
    
    public void focusGained(FocusEvent e) {
        JTextField source = (JTextField) e.getSource();
        source.setSelectionStart(0);
        source.setSelectionEnd(source.getText().length());
    }
    
    public void focusLost(FocusEvent e) {
        JTextField source = (JTextField) e.getSource();
        try {
            String[] args = source.getText().split("[, ]");
            for (int i = 0; i < args.length; ++i) 
                if (!args[i].equals("")) {
                    double d = Double.parseDouble(args[i]);
                    if ((source == lambdaField) && (d < 0))
                        source.setText(Messages.ERROR_MUST_BE_POSITIVE);
                }
        } catch (NumberFormatException ex) {
            source.setText(Messages.ERROR_COMMA_LIST);
        }
    }
    
    
    /**Sets all the options on this panel. All arrays are booleans for specific buttons on the panel.
     * @param normalizes [Normalize true?, Normalize false?]
     * @param kernels [Direct?, Linear?, Polynomial?, RBF?, Cosine?]
     * @param polyParams The set of parameters for the polynomial kernel
     * @param rbfParams The set of parameters for the RBF kernel
     * @param bias The set of biases to use
     * @param priors [Laplacian?, Gaussian?, None?]
     * @param lambdas The set of lambdas.
     * @param updateRules [Component-wise?, Non-component-wise?]
     */
    public void setOptions(boolean[] normalizes, boolean[] kernels, TreeSet polyParams, TreeSet rbfParams, TreeSet biases, boolean[] priors, TreeSet lambdas, boolean[] updateRules) {
        cbNormalizeYes.setSelected(normalizes[0]);
        cbNormalizeNo.setSelected(normalizes[1]);
        
        for (int i = 0; i < kernels.length; ++i)
            cbKernels[i].setSelected(kernels[i]);
        
        String s = "";
        ArrayList pParams = new ArrayList();
        pParams.addAll(polyParams);
        for (int i = 0; i < polyParams.size(); ++i) {
            s += pParams.get(i);
            if (i != pParams.size()-1) s += ", ";
        }
        polyParamField.setText(s);
        
        s = "";
        ArrayList rParams = new ArrayList();
        rParams.addAll(rbfParams);
        for (int i = 0; i < rbfParams.size(); ++i) {
            s += rParams.get(i);
            if (i != rParams.size()-1) s += ", ";
        }
        rbfParamField.setText(s);
        
        if (biases.contains(new Double(0))) cbBiasNo.setSelected(true);
        
        s = "";
        ArrayList b = new ArrayList();
        b.addAll(biases);
        for (int i = 0; i < biases.size(); ++i) {
        	if (((Double) b.get(i)).doubleValue() == 0) continue;
        	else cbBiasYes.setSelected(true);
        	s += b.get(i);
            if (i != b.size()-1) s += ", ";
        }
        biasField.setText(s);
        
        for (int i = 0; i < priors.length; ++i)
            cbPriors[i].setSelected(priors[i]);
        
        s = "";
        ArrayList l = new ArrayList();
        l.addAll(lambdas);
        for (int i = 0; i < lambdas.size(); ++i) {
            s += l.get(i);
            if (i != l.size()-1) s += ", ";
        }
        lambdaField.setText(s);
        
        cbStep.setSelected(updateRules[0]);
        cbIter.setSelected(updateRules[1]);
    }
}
