package edu.duke.cs.smlr.learn;


import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeSet;

import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.learn.properties.ComboProperty;
import edu.duke.cs.smlr.learn.properties.KFoldProperty;
import edu.duke.cs.smlr.learn.properties.LearnProperty;

import edu.duke.cs.smlr.util.Constants;
import edu.duke.cs.smlr.util.DataPointMap;
import edu.duke.cs.smlr.util.Defaults;

/**
 * Class to take a data set and make the K appropriate test sets
 * 
 * @author Matt Edwards, Jason Bosko
 */

public class KFolder {

	private static int rand(int max) {
		return (int) (Math.random() * Integer.MAX_VALUE) % max;
	}

	/**
	 * Generates an arraylist of rows from 0..<b>rowCount</b> that are not contained in <b>rows</b>
	 * @param rows The rows that should not be included.
	 * @param rowCount The total number of rows.
	 * @return The arraylist of "other" rows.
	 */
	private static ArrayList otherRows(ArrayList rows, int rowCount) {
		ArrayList ans = new ArrayList();
		boolean[] hit = new boolean[rowCount];
		for (Iterator it = rows.iterator(); it.hasNext();)
			hit[((Integer) it.next()).intValue()] = true;
		for (int i = 0; i < hit.length; ++i)
			if (!hit[i])
				ans.add(new Integer(i));
		return ans;
	}

	/**
	 * Generates the comboproperty specified by the rows in the test set (<b>rows</b>) and the columns in the test 
	 * set (<b>cols</b>), assuming that <b>m1</b> is the features of the  data and <b>m2</b> is the classifications.
	 * @param ans The comboproperty to put the information in.
	 * @param m1 The features of the data
	 * @param m2 The classifications of the data
	 * @param rows The rows to be excluded from the training set.
	 * @param cols Usually this is just every column.
	 */
	private static void process(ComboProperty ans, IMatrix m1, IMatrix m2,
			ArrayList rows, ArrayList cols) {
		//ok this has to be views here
		ans.getTestProperty().setTestData(m1.chunkView(rows, cols));
		ans.getLearnProperty().setTrainingData(
				m1.chunkView(otherRows(rows, m1.numRows()), cols));
		ans.getTestProperty().setTrainingData(
				m1.chunkView(otherRows(rows, m1.numRows()), cols));
		ArrayList temp = new ArrayList();
		temp.add(new Integer(0));
		ans.getTestProperty().setClassifications(m2.chunkView(rows, temp));
		ans.getLearnProperty().setClassifications(
				m2.chunkView(otherRows(rows, m2.numRows()), temp));
	}

	/**
	 * Creates the full set of included/excluded data pairs specified by the options in <b>prop</b>
	 * @param prop Contains the options for cross-validation
	 * @return The full set of included/excluded data pairs.
	 */
	public static ComboProperty[] kfold(KFoldProperty prop) {
		return kfold(prop.getLearnProperty(), prop.getK(), prop.getBehavior());
	}

	/**
	 * Creates the full set of included/excluded data pairs specified by <b>props</b>, <b>k</b>, and <b>behavior</b>.
	 * @param props The learn property with the data and the options that remain constant across all folds.
	 * @param k Specifies how many folds to make. If k=1, then the default of k=5 is used.
	 * @param behavior Specifies how to choose the data points left out of each fold.
	 * @return The full set of included/excluded data pairs.
	 */
	public static ComboProperty[] kfold(LearnProperty props, int k, int behavior) {
		// assuming that a valid behavior was given

		if (k == 1)
			k = Defaults.k; //this is our default k if none is given
		
		ComboProperty[] ans = new ComboProperty[k];
		for (int i = 0; i < k; ++i)
			ans[i] = new ComboProperty(props);
		IMatrix x = props.getTrainingData();
		IMatrix y = props.getClassifications();

		boolean[] used = new boolean[x.numRows()];
		TreeSet sortedTestRows = new TreeSet();
		ArrayList testRows = new ArrayList();
		ArrayList allCols = new ArrayList();
		for (int i = 0; i < x.numCols(); ++i)
			allCols.add(new Integer(i));

		// this part is the LOOCV
		if (k == x.numRows()) {
			for (int i = 0; i < k; i++) {
				DataPointMap trainMap = new DataPointMap();
				DataPointMap testMap = new DataPointMap();
				testRows.clear();
				testRows.add(new Integer(i));
				process(ans[i], x, y, testRows, allCols);
				for (int c = 0; c < k; c++) {
					if (c == i)		testMap.put(new Integer(i), new Integer(0));
					else if (c < i)	trainMap.put(new Integer(c), new Integer(c));
					else if (c > i) trainMap.put(new Integer(c), new Integer(c-1));
				}
				ans[i].setLearnDataMap(trainMap);
				ans[i].setTestDataMap(testMap);
			}
			return ans;
		}
		// end of LOOCV

		if (behavior == Constants.KFOLD_RATIO_RANDOM) {
			// ratio random, make k - 1 tests
			// TODO: make this not break for sketchy input
			int classes = y.numCols();
			if (classes == 1)
				classes = (int) y.maxAbsRowSum();
			int thingsPerClass[] = new int[classes];
			int extrasPerClass[] = new int[classes];
			for (int i = 0; i < y.numRows(); ++i)
				thingsPerClass[(int) y.get(i, 0) - 1]++;
			int maxExtra = 0;
			for (int i = 0; i < classes; ++i) {
				extrasPerClass[i] = thingsPerClass[i] % k;
				if (extrasPerClass[i] > maxExtra) maxExtra = extrasPerClass[i];
				thingsPerClass[i] /= k;
			}
			int total = 0;
			for (int i = 0; i < classes; ++i)
				total += thingsPerClass[i];
			int leftover = x.numRows() / k - total;
			// if(leftover != 0) System.err.println("leftover = " + leftover);
			int curr;
			for (int i = 0; i < k - 1; ++i) {
				testRows.clear();
				sortedTestRows.clear();
				for (int j = 0; j < classes; ++j)
					for (int m = 0; m < thingsPerClass[j]; ++m) {
						do
							curr = rand(x.numRows());
						while (used[curr] || ((int) y.get(curr, 0) != j + 1));
						used[curr] = true;
						sortedTestRows.add(new Integer(curr));
					}
				int extraTest = leftover;
				if (i < x.numRows() % k) extraTest++;
				for (int j = 0; j < extraTest; ++j) {
					do
						curr = rand(x.numRows());
					while (used[curr] || extrasPerClass[(int) y.get(curr, 0) - 1] != maxExtra || extrasPerClass[(int) y.get(curr, 0) - 1] <= 0);
					used[curr] = true;
					extrasPerClass[(int) y.get(curr, 0)-1]--;
					sortedTestRows.add(new Integer(curr));
					maxExtra--;
					for (int a = 0; a < extrasPerClass.length; ++a) 
						if (extrasPerClass[a] > maxExtra) maxExtra = extrasPerClass[a];
				}
				// System.out.print("[");
				// for(int q = 0; q < used.length; ++q)
				// if(used[q]) System.out.print("*");
				// else System.out.print((int) y.get(q, 0));
				// System.out.println("]");
				testRows.addAll(sortedTestRows);
				process(ans[i], x, y, testRows, allCols);
				DataPointMap trainMap = new DataPointMap();
				DataPointMap testMap = new DataPointMap();
				int trainRow = 0;
				int testRow = 0;
				for (int c = 0; c < x.numRows(); c++) {
					if (sortedTestRows.contains(new Integer(c))) {
						testMap.put(new Integer(c), new Integer(testRow));
						testRow++;
					}
					else {
						trainMap.put(new Integer(c), new Integer(trainRow));
						trainRow++;
					}
				}
				ans[i].setLearnDataMap(trainMap);
				ans[i].setTestDataMap(testMap);
			}
		} else {
			// pure random or non-random, make k - 1 tests

			for (int i = 0; i < k - 1; ++i) {
				testRows.clear();
				sortedTestRows.clear();
				int numTest = x.numRows() / k;
				if (i < x.numRows() % k) numTest++;
				for (int j = 0; j < numTest; ++j) {
					int curr;
					if (behavior == Constants.KFOLD_PURE_RANDOM)
						curr = rand(x.numRows());
					else
						curr = 0;
					while (used[curr])
						if (behavior == Constants.KFOLD_PURE_RANDOM)
							curr = rand(x.numRows());
						else
							curr++;
					used[curr] = true;
					sortedTestRows.add(new Integer(curr));
				}
				testRows.addAll(sortedTestRows);
				process(ans[i], x, y, testRows, allCols);
				DataPointMap trainMap = new DataPointMap();
				DataPointMap testMap = new DataPointMap();
				int trainRow = 0;
				int testRow = 0;
				for (int c = 0; c < x.numRows(); c++) {
					if (sortedTestRows.contains(new Integer(c))) {
						testMap.put(new Integer(c), new Integer(testRow));
						testRow++;
					}
					else {
						trainMap.put(new Integer(c), new Integer(trainRow));
						trainRow++;
					}
				}
				ans[i].setLearnDataMap(trainMap);
				ans[i].setTestDataMap(testMap);
			}
		}
		// make the k-th test from all unused elements

		testRows.clear();
		for (int i = 0; i < used.length; ++i)
			if (!used[i])
				testRows.add(new Integer(i));
		process(ans[k - 1], x, y, testRows, allCols);
		DataPointMap trainMap = new DataPointMap();
		DataPointMap testMap = new DataPointMap();
		int trainRow = 0;
		int testRow = 0;
		for (int c = 0; c < x.numRows(); c++) {
			if (testRows.contains(new Integer(c))) {
				testMap.put(new Integer(c), new Integer(testRow));
				testRow++;
			}
			else {
				trainMap.put(new Integer(c), new Integer(trainRow));
				trainRow++;
			}
		}
		ans[k-1].setLearnDataMap(trainMap);
		ans[k-1].setTestDataMap(testMap);
		return ans;
	}
}
