package edu.duke.cs.smlr.learn;

import edu.duke.cs.smlr.factories.KernelFactory;
import edu.duke.cs.smlr.learn.interfaces.IKernel;
import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.learn.kernels.AbstractKernel;
import edu.duke.cs.smlr.learn.properties.LearnProperty;
import edu.duke.cs.smlr.util.DataPointMap;

/**
 * Stores the information needed to cache kernel functions in order to 
 * reduce the time spent in cross-validation.
 * 
 * @author Matt Edwards
 *
 */
public class KernelSource {
	IMatrix everything;
	
	/**
	 * Constructs a KernelSource from the data inside a <b>lp</b>
	 * @param lp The LearnProperty with the kernel information to be stored
	 */
	public KernelSource(LearnProperty lp) {
		IKernel kernel = KernelFactory.makeKernel(lp.getKernel());
		everything = kernel.computeKernel(lp.getTrainingData(),
				lp.getTrainingData(), lp.getKernelPar(), lp
				.getBias(), lp.isNormalize());
		//System.err.println("Caching the pairwise kerneling of " + lp.getTrainingData().numRows() + " examples");
		//System.err.println("Producing a " + everything.numRows() + " by " + everything.numCols() + " matrix");
	}
	
	/**
	 * Produces kerneled data by pulling the appropriate entries out of the 
	 * cached kernel source
	 * @param test The map of test data points to their position in the original matrix
	 * @param train The map of train data points to their position in the original matrix
	 * @return The kerneled data, obtained through the cache
	 */
	public IMatrix makeKernel(DataPointMap test, DataPointMap train, double bias) {
		int m = test.getNewToOld().size();
		int n = train.getNewToOld().size();
		IMatrix ret = everything.like(m, n);
		for (int i = 0; i < m; ++i)
			for (int j = 0; j < n; ++j)
				ret.set(everything.get(test.getOriginalDataPoint(new Integer(i)), 
						train.getOriginalDataPoint(new Integer(j))), 
						i, j);
		if (everything.numRows() != everything.numCols())
			ret = AbstractKernel.addBias(ret, bias);
		return ret;
	}
	
	/**
	 * Clears the kernel source cache
	 *
	 */
	public void clear() {
		everything = null;
	}
}
