package edu.duke.cs.smlr.learn;

import edu.duke.cs.smlr.factories.KernelFactory;
import edu.duke.cs.smlr.factories.LearnerFactory;
import edu.duke.cs.smlr.learn.interfaces.IKernel;
import edu.duke.cs.smlr.learn.interfaces.ILearner;
import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.learn.properties.LearnProperty;
import edu.duke.cs.smlr.learn.results.LearnResult;
import edu.duke.cs.smlr.util.DataPointMap;

/**
 * Class that learns a set of weights, effectively converting a LearnProperty
 * class into a LearnResult
 * 
 * @author Matt Edwards, Jason Bosko
 */
public class Learner {
	private Learner () {} // no instantiation allowed
	
	/**
	 * Learns a classifier from the options in <b>settings</b>.
	 * @param settings The options for learning.
	 * @param map The map of data points in the array to their actual position in the original data.
	 * @param cache The kernel cache (should be <b>null</b> if no cache is needed/wanted.
	 * @return The result of the learning.
	 */
	public static LearnResult learn(LearnProperty settings, DataPointMap map, KernelSource cache) {
		IMatrix kerneledData = null;//, kerneledData1 = null;
		if (cache != null) {
			kerneledData  = cache.makeKernel(map, map, settings.getBias());			
		} 
		else 
		{
			IKernel kernel = KernelFactory.makeKernel(settings.getKernel());
			kerneledData = kernel.computeKernel(settings.getTrainingData(),
					settings.getTrainingData(), settings.getKernelPar(), settings
					.getBias(), settings.isNormalize());
		}
		/*if (kerneledData.sum() != kerneledData1.sum()) {
			System.out.println(kerneledData.numRows() + " by " + kerneledData.numCols());
			System.out.println(kerneledData.chunk(0, 0, 3, 25));
			System.out.println(kerneledData1.numRows() + " by " + kerneledData1.numCols());			
			System.out.println(kerneledData1.chunk(0, 0, 3, 25));
			System.out.println(kerneledData.termDiv(kerneledData1).chunk(0, 0, 3, 25));
			"".charAt(4);
		}
		else
			System.err.println("Cached kernel evaluations in learn are correct!");
		*/
		ILearner learner = LearnerFactory.makeLearner(settings.getMethod());
		
		double startTime = System.currentTimeMillis();
		LearnResult results = learner.learnWeights(kerneledData, settings
				.getClassifications(), settings.getLambda(), settings
				.getConvTol(), settings.getKillThreshold(), null, settings
				.getPrior(), settings.isUseOverrelaxation(), settings.isCacheExp(),
				settings.getMaxIters());
		double endTime = System.currentTimeMillis();
		
		int basisFunctions = 0;
		IMatrix weights = results.getWeights();
		for (int i = 0; i < weights.numRows(); i++) {
			boolean basisUsed = false;
			for (int j = 0; j < weights.numCols(); j++) {
				if (Math.abs(weights.get(i, j)) > 0)
					basisUsed = true;
			}
			if (basisUsed) basisFunctions++;
		}
		results.setBasisUsed(basisFunctions);
		results.setTime((endTime - startTime) / 1000);
		
		return results;
	}
}
