package edu.duke.cs.smlr.learn.classifiers;

import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.learn.results.ROCResults;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * This class takes a set of probabilities from an entire run and varies the 
 * threshold for classification in order to generate an ROC.
 * 
 * @author Jason Bosko
 */
public class ClassifierThresholdFinder {
	/**
	 * Generates the data for an ROC
	 * @param probs The matrix of probabilities computed by the classifier for each point
	 * @param classes The real classifications associated with the data
	 * @return An ROCResults that contains the ROC data for the given probabilities.
	 */
	public ROCResults generateROC(IMatrix probs, IMatrix classes, int posCol) {
		if (posCol < 1 || posCol > 2) posCol = 2;
		int totalPos = 0, totalNeg = 0, total = classes.numRows();
		double posP[] = new double[total];
		double negP[] = new double[total];
		
		// Turn the matrix of probabilities into two sorted arrays, one for the 
		// probabilites of being positive for actual positive points and one for
		// the probabilities of being positive for actual negative points.
		for (int i = 0; i < total; i++) {
			if ((int) classes.getQuick(i,0) == posCol) {
				posP[totalPos] = probs.getQuick(i, posCol-1);
				totalPos++;
			}
			else {
				negP[totalNeg] = probs.getQuick(i, posCol-1);
				totalNeg++;
			}
		}
		Arrays.sort(posP,0,totalPos); Arrays.sort(negP,0,totalNeg);
		
		// Starting with a threshold of 0 (everything classified as positive), 
		// iterate through all values of probabilities for positive points, 
		// each time losing one true positive, and checking how many 
		// false positives are lost
		ArrayList threshes = new ArrayList();
		ArrayList trues = new ArrayList();
		ArrayList falses = new ArrayList();
		double thresh = 0;
		int numNeg = totalNeg, numPos = totalPos;
		// Initial threshold of 0
		threshes.add(new Double(thresh));
		trues.add(new Integer(numPos));
		falses.add(new Integer(numNeg));
		numPos++;
		// Start with the threshold defined by the lowest probability for a 
		// true positive point, and use the next probability of a true 
		// positive point and so on until no more positive points are left
		while (numPos > 1) {
			numPos--;
			thresh = posP[totalPos-numPos];
			while (numPos >= 0 && posP[totalPos-(numPos-1)]==thresh) numPos--;
			while (numNeg > 0 && negP[totalNeg-numNeg] < thresh) numNeg--;
			threshes.add(new Double(thresh));
			trues.add(new Integer(numPos));
			falses.add(new Integer(numNeg));
		}
		
		// Convert all the ArrayLists into arrays
		int s = threshes.size();
		double th[] = new double[s];
		int tp[] = new int[s];
		int fp[] = new int[s];
		double tpr[] = new double[s];
		double fpr[] = new double[s];
		for (int i = 0; i < s; i++) {
			th[i] = ((Double) threshes.get(i)).doubleValue();
			tp[i] = ((Integer) trues.get(i)).intValue();
			fp[i] = ((Integer) falses.get(i)).intValue();
			tpr[i] = (double) tp[i] / (double) totalPos;
			fpr[i] = (double) fp[i] / (double) totalNeg;
		}
		
		// Create the ROCResults
		ROCResults roc = new ROCResults(th, tp, fp, tpr, fpr, totalPos, totalNeg);
		
		// Compute the threshold defined by the ratio of positive data to the 
		// entire data, and find the true positive and false positive counts
		double opthresh = (double) totalPos / (double) total;
		numPos = 0; numNeg = 0;
		for (int i = 0; i < totalPos; i++) {
			if (posP[i] >= opthresh) numPos++;
		}
		for (int i = 0; i < totalNeg; i++) {
			if (negP[i] >= opthresh) numNeg++;
		}
		roc.setOptimumRatio(opthresh, numPos, numNeg);
		
		
		return roc;
	}
}
