package edu.duke.cs.smlr.learn.interfaces;

import java.util.AbstractList;

import edu.duke.cs.smlr.util.DataPointMap;
import edu.duke.cs.smlr.util.matrix.AbstractMatrix;
import edu.duke.cs.smlr.util.matrix.ManipMatrix;
import edu.duke.cs.smlr.util.matrix.MathMatrix;
import edu.duke.cs.smlr.util.matrix.SimpleMatrix;

/**This interface contains all methods that must be implemented to have a
 * "full" matrix implementation in the sense of this program.  Its generic
 * implementation, using methods that can be written only by using other 
 * methods in the interface, is accomplished by a series of abstract classes.
 * The abstract classes are split up, more or less based on the type of methods
 * they implement, to improve readability and also to allow focus for speed 
 * improvements (that is, hypothetically only a small subset of the abstract 
 * classes will need to be overridden to speed up a particular task).
 * 
 * @see AbstractMatrix
 * @see SimpleMatrix
 * @see MathMatrix
 * @see ManipMatrix
 * 
 * @author Matt Edwards
 */

public interface IMatrix extends Cloneable, java.io.Serializable {

    /**Returns the number of rows in the matrix
     * @return the number of rows in the matrix
     */
    public int numRows();

    /**Returns the number of columns in the matrix
     * @return the number of columns in the matrix
     */
    public int numCols();

    /**Gets a double from the specified position in the matrix
     * <br>
     * The indices will not be verified as in range by the method; if the
     * indices are out of bounds, undefined behavior will result
     * @param r the row in the matrix
     * @param c the column in the matrix
     * @return the value at the specified position
     */
    public double getQuick(int r, int c);

    /**Sets a double at the specified position in the matrix
     * <br>
     * The indices will not be verified as in range by the method; if the
     * indices are out of bounds, undefined behavior will result
     * @param d the value to set
     * @param r the row in the matrix
     * @param c the column in the matrix
     */
    public void setQuick(double d, int r, int c);

    /**Gets a double from the specified position in the matrix
     * <br>
     * The indices given will be verified as legal by the method
     * @param r the row in the matrix
     * @param c the column in the matrix
     * @return the value at the specified position
     * @throws IndexOutOfBoundsException if the indices are out of bounds
     */
    public double get(int r, int c) throws IndexOutOfBoundsException;

    /**Sets a double at the specified position in the matrix
     * <br>
     * The indices given will be verified as legal by the method
     * @param d the value to set
     * @param r the row in the matrix
     * @param c the column in the matrix
     * @throws IndexOutOfBoundsException if the indices are out of bounds
     */
    public void set(double d, int r, int c)
            throws IndexOutOfBoundsException;

    /**Returns a zero-filled matrix of the given size of the same dynamic type as <b>this</b>
     * @param r the number of rows in the new matrix
     * @param c the number of columns in the new matrix
     * @return a matrix of the same dynamic type
     * @throws IllegalArgumentException if r or c are negative
     */
    public IMatrix like(int r, int c)
    		throws IllegalArgumentException;

    /**Returns a zero-filled matrix of the same size and dynamic type as <b>this</b>
     * @return a matrix of the same dynamic type
     */
    public IMatrix like();

    /**Replaces the elements of <b>this</b> with the values from another matrix
     * <br>
     * The two matrices must be of the same dimension
     * @param other the matrix to copy from, unaltered
     * @throws IllegalArgumentException if the matrix dimensions do not match
     */
    public void assign(IMatrix other)
            throws IllegalArgumentException;

    /**Returns the matrix as a two-dimensional double array.  The returned array
     * should be row major, that is, it should be accessed as array[row][column].
     * The array is not backed by the matrix.
     * @return a two-dimensional double array of the matrix values
     */
    public double[][] toArray();
    
    /**Returns the maximum absolute valued row sum
     * @return the maximum absolute valued row sum
     */
    public abstract int maxAbsRowSum();

    /**Returns a column vector of row sums
     * @return a column vector of row sums
     */
    public abstract IMatrix rowSum();

    /**Returns a row vector of column sums
     * @return a row vector of column sums
     */
    public abstract IMatrix columnSum();

    /**Returns the sum of every element in the matrix
     * @return the sum of every element in the matrix
     */
    public abstract double sum();

    /**Returns a row vector of the averages of each column
     * @return a row vector of the averages of each column
     */
    public abstract IMatrix columnMean();

    /**Returns a row vector of the standard deviations of each column
     * @return a row vector of the standard deviations of each column
     */
    public abstract IMatrix columnStd();

    /**Returns a column vector of the row index of the maximum absolute value in each row
     * <br>
     * In the case of a tie, the lowest index (first occurrence) is used
     * @return a column vector of row indices
     */
    public abstract IMatrix maxIndices();

    /**Returns a column vector of the maximum absolute value in each row
     * @return a column vector of maximum absolute values
     */
    public abstract IMatrix maxEachRow();

    /**Returns the sum of absolute values of elements in the matrix
     * @return the sum of absolute values
     */
    public abstract double sumAbs();

    /**Multiplies all elements by a scalar
     * @param d the scalar to multiply by
     * @return <b>this</b> for convenience
     */
    public abstract IMatrix mult(double d);

    /**Adds to all elements
     * @param d the scalar to add
     * @return <b>this</b> for convenience
     */
    public abstract IMatrix add(double d);

    /**Raises each element to a power
     * @param d the power to raise each element to
     * @return <b>this</b> for convenience
     */
    public abstract IMatrix termPow(double d);

    /**Exponentiates each element
     * @return <b>this</b> for convenience
     */
    public abstract IMatrix exp();

    /**Replaces each element with its natural logarithm
     * @return <b>this</b> for convenience
     */
    public abstract IMatrix log();

    /**Computes the Euclidean two-norm if this matrix is a row or column vector
     * <br>
     * In general, this will calculate the square root of the sum of the squares of each element in the
     * matrix
     * @return the square root of the sum of the squares of each element
     */
    public abstract double norm();
    
    /**Inverses the matrix. This method alters <b>this</b>.
     * @return <b>this</b> for convenience.
     * 
     * @throws UnsupportedOperationException
     */
    public abstract IMatrix inverse() throws UnsupportedOperationException;

    /**Transposes this matrix
     * <br>
     * The transposition is reflected in the matrix, and subsequent use of the transposed
     * matrix will change the original matrix.
     * @return <b>this</b>, for convenience     * 
     */
    public abstract IMatrix transpose();

    public abstract IMatrix mult(IMatrix other)
            throws IllegalArgumentException;

    /**Adds another matrix to this matrix
     * <br>
     * This method alters <b> this </b>
     * @param other the matrix to add, unaltered
     * @return <b>this</b> for convenience
     * @throws IllegalArgumentException if the matrix dimensions do not match
     */
    public abstract IMatrix add(IMatrix other)
            throws IllegalArgumentException;

    /**Multiplies another matrix with this matrix, term-by-term
     * <br>
     * This method alters <b> this </b>
     * @param other the matrix to multiply with, unaltered
     * @return <b>this</b> for convenience
     * @throws IllegalArgumentException if the matrix dimensions do not match
     */
    public abstract IMatrix termMult(IMatrix other)
            throws IllegalArgumentException;

    /**Divides this matrix by another matrix, term-by-term
     * <br>
     * This method alters <b>this</b>
     * @param other the matrix to divide by
     * @return <b>this</b> for convenience
     * @throws IllegalArgumentException if the matrix dimensions do not match
     */
    public abstract IMatrix termDiv(IMatrix other)
            throws IllegalArgumentException;

    /**Computes the Kronecker product between this and another matrix
     * <br>
     * This method does not alter either matrix
     * @param other the matrix to compute the Kronecker product with
     * @return the Kronecker product of the two matrices
     */
    public abstract IMatrix kronecker(IMatrix other);    
    
    /**Returns a chunk of a certain part of the matrix
     * <br>
     * This method does not alter <b> this </b>
     * <br>
     * The returned chunk is not backed by the original matrix
     * @param startRow the row index to start at (inclusive)
     * @param startCol the column index to start at (inclusive)
     * @param endRow the row index to stop at (exclusive)
     * @param endCol the column index to stop at (exclusive)
     * @return a matrix of the specified chunk
     * @throws ArrayIndexOutOfBoundsException if any indices given are out of bounds
     */
    public abstract IMatrix chunk(int startRow, int startCol, int endRow,
            int endCol) throws ArrayIndexOutOfBoundsException;

    /**Returns a chunk of the matrix based on lists of indices
     * <br>
     * This method does not alter the original matrix
     * @param rowIndices a list of row indices to use
     * @param colIndices a list of column indices to use
     * @return the specified chunk
     * @throws ArrayIndexOutOfBoundsException if either list contains an index that is out of bounds
     */
    public abstract IMatrix chunk(AbstractList rowIndices,
            AbstractList colIndices) throws ArrayIndexOutOfBoundsException;

    /**Constructs a block-matrix of the original matrix, Matlab (r) repmat style
     * <br>
     * This method does not alter the original matrix
     * @param r number of rows in the block matrix
     * @param c number of columns in the block matrix
     * @return the new block matrix
     * @throws IllegalArgumentException if the arguments are less than one
     */
    public abstract IMatrix repmat(int r, int c)
            throws IllegalArgumentException;

    /**Transforms <b>this</b>, a row or column vector, into a square matrix, with the main
     * diagonal the elements of the row or column matrix
     * <br>
     * This method does not alter the original method
     * <br>
     * If <b>this</b> is not a row or column vector then this behaves the same as clone
     * @return a diagonal matrix or a deep copy
     */
    public abstract IMatrix diag();

    /**Changes the dimensions of <b>this</b> by reshaping it in a row-major form
     * <br>
     * This method does not alter the original matrix
     * @param r the number of rows in the new matrix
     * @param c the number of columns in the new matrix
     * @return a matrix of the given size
     * @throws IllegalArgumentException if the new dimensions are invalid
     */
    public abstract IMatrix reshape(int r, int c);

    /**TODO What the heck does this do?
     * @param indices
     * @param maxIndex
     * @return an expanded matrix
     */
    public abstract IMatrix expand(AbstractList indices, int maxIndex);
    
    /**Returns a deep copy of the matrix.  The returned matrix is not backed by
     * the matrix.
     * @return a clone of the matrix
     */
    public abstract Object clone();
    
    /**Returns a deep copy of the matrix.  Same as clone, but the return type is
     * an IMatrix to eliminate a cast.
     * 
     * @see IMatrix#clone()
     * 
     * @return a clone of the matrix
     */
    public abstract IMatrix copy();

	/**Creates a chunk like a call to chunk, but does not duplicate storage.
	 * Returned chunk is backed by the original matrix.
	 * 
	 * @param rows
	 * @param temp
	 * @return a chunk
	 */
	public IMatrix chunkView(AbstractList rows, AbstractList temp);
	
	/**
	 * Prints the matrix, using the map of indices to their original index in an 
	 * original matrix.
	 * @param map The map of array indices to original indices.
	 * @return The matrix as a string
	 */
	public String toString(DataPointMap map);
}
