package edu.duke.cs.smlr.learn.kernels;

import edu.duke.cs.smlr.factories.MatrixFactory;
import edu.duke.cs.smlr.learn.interfaces.IKernel;
import edu.duke.cs.smlr.learn.interfaces.IMatrix;

/**An abstract kernel class containing methods used for all kernels.
 * @author Matt Edwards
 *
 */
public abstract class AbstractKernel implements IKernel {

	public abstract IMatrix computeKernel(IMatrix X1, IMatrix X2, Object kernelPar, double bias, boolean normalize);
	
	/**Normalizes the test data with the equations that make the training data have a mean of zero and a unit variance
	 * @param test the test data to normalize, unaltered
	 * @param train the training data to get the normalization formulas from, unaltered
	 * @return normalized testing data
	 */
	protected static IMatrix normalizeTest(IMatrix test, IMatrix train) {
		int xrows = test.numRows();
		int xcols = test.numCols();

		IMatrix avg = train.columnMean();
		IMatrix dev = train.columnStd();
		
		for (int k = 0; k < xcols; k++)
			if (dev.get(0, k) == 0)	dev.set(1.0, 0, k);

		test = test.copy().add(avg.repmat(xrows, 1).mult(-1.0)).termDiv(
				dev.repmat(xrows, 1));
		
		return test;
	}
	
	/**Normalizes the original training data to have a mean of zero and a unit variance
	 * @param test the test data, unused, included to provide symmetry with normalizeTest
	 * @param train the training data to normalize, unaltered
	 * @return normalized training data
	 */
	protected static IMatrix normalizeTrain(IMatrix test, IMatrix train) {
		int xrows = train.numRows();
		int xcols = train.numCols();

		IMatrix avg = train.columnMean();
		IMatrix dev = train.columnStd();
		for (int k = 0; k < xcols; k++)
			if (dev.get(0, k) == 0)
				dev.set(1.0, 0, k);

		train = train.copy().add(avg.repmat(xrows, 1).mult(-1.0)).termDiv(
				dev.repmat(xrows, 1));
		
		return train;		
	}
	
	/**Adds a column of ones to a matrix
	 * @param m the original
	 * @return a deep copy of the original matrix with a column of ones added as the last column
	 */
	public static IMatrix addBias(IMatrix m, double bias) {
		if (bias == 0) return m;
		int ansRows = m.numRows();
		int ansCols = m.numCols() + 1;
		IMatrix ans = MatrixFactory.makeMatrix(ansRows, ansCols);
		for (int k = 0; k < ansRows; ++k)
			for (int i = 0; i < ansCols; ++i)
				if (i == ansCols - 1)
					ans.set(1, k, i);
				else
					ans.set(m.get(k, i), k, i);
		return ans;
	}

}
