package edu.duke.cs.smlr.learn.learners;

import edu.duke.cs.smlr.factories.MatrixFactory;
import edu.duke.cs.smlr.learn.interfaces.ILearner;
import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.learn.results.LearnResult;
import edu.duke.cs.smlr.util.FastRandom;
import edu.duke.cs.smlr.util.Constants;

/**An optimized version of the component-wise SMLR learning algorithm.  This
 * features several optimizations, including no object creation or function
 * calls in the inner loop.  It is based on using primitive double arrays to store
 * the data, since this algorithm requires no complex matrix operations.
 * 
 * @author Matt Edwards, Jason Bosko
 *
 */
public class ArrayStepwiseLearner extends AbstractLearner implements ILearner {
    
    public LearnResult learnWeights(IMatrix xMatrix, IMatrix y,
            double lambda, double convergenceTol, double killThreshold,
            IMatrix startAt, int method, boolean overrelaxed, final boolean cacheExp, int maxIterations) {
   	
        y = processClasses(y);

        
        int n = y.numRows();
        int m = y.numCols();
        int d = xMatrix.numCols();
        
        m--;
        IMatrix xx = xMatrix.copy().termMult(xMatrix); 
        IMatrix autoCorrMatrix = xx.columnSum().mult((double) (m) / (double) (2.0 * (m + 1))).transpose(); //column vector
        IMatrix xyMatrix = xMatrix.copy().transpose().mult(y.chunk(0, 0, y.numRows(), m));
        
        double[] autoCorr = new double[autoCorrMatrix.numRows()];
        for(int i = 0; i < autoCorr.length; ++i) autoCorr[i] = autoCorrMatrix.get(i, 0);
        
        double[] xy = new double[d * m];
        for(int i = 0; i < xyMatrix.numRows(); ++i)
            for(int j = 0; j < xyMatrix.numCols(); ++j)
                xy[i * m + j] = xyMatrix.get(i, j);
        double[][] xTranspose = xMatrix.copy().transpose().toArray();		
        double[] w = new double[d * m];
        if(startAt != null) w = startAt.reshape(1, d * m).toArray()[0];
        double[] wPrev = new double[d * m];
        if(startAt != null) wPrev = startAt.reshape(1, d * m).toArray()[0];        
        double[][] xwTranspose = new double[m][n];
        double[][] Etranspose = new double[m][n]; 
        double[] eNewM = new double[n];
        for(int i = 0; i < n; ++i)
            for(int j = 0; j < m; ++j)
                Etranspose[j][i] = 1;
        
        double[] S = new double[n];
        for(int i = 0; i < n; ++i) S[i] = m + 1;
        
        boolean converged = false;
        int l = 0;
        double incr = 1e10;
        int b = 0;
        int mm = 0;
        double wNew = 0;
        double wOld = 0;
        double testZeroBasis = 1;
        boolean changed = false;
        double nonZero = 0;
        double grad = 0;
        int dTimesm = d * m;
        int index = 0;
        double[] tempArray, tempArray2;
        //let's allow GC to reclaim this
        xx = null;
        autoCorrMatrix = null;
        xyMatrix = null;
        xMatrix = null;
        y = null;
        startAt = null;
        //System.gc();
        double temp, temp2, numer, denom, dTimesmDouble = dTimesm;
        int i;
        lambda /= 2.0;
        
        
        final boolean PRECOMPUTE_EXP = cacheExp;
        double[] table1 = null;
        double[] table2 = null;
        if(PRECOMPUTE_EXP) {
            table1 = new double[2000];
            for(i = 0; i < 2000; ++i) table1[i] = Math.exp(((double) i / 100.0) - 20.0);
            table2 = new double[10000];
            for(i = 0; i < 10000; ++i) table2[i] = Math.exp((double) i / 1000.0);
        }
        
        while (!converged) {
            index = b * m + mm;
            wOld = w[index];
            if (wOld != 0 || FastRandom.random() <= testZeroBasis) {
                grad = xy[index];
                tempArray = xTranspose[b];
                tempArray2 = Etranspose[mm];
                for(i = 0; i < n; ++i) grad -= tempArray[i] * tempArray2[i] / S[i];
                if (method == Constants.PRIOR_GAUSSIAN) {
                	wNew = ((wOld * autoCorr[b]) + grad) / (autoCorr[b] + lambda*2);
                	changed = true;
                }
                else if (method == Constants.PRIOR_NONE) { 
                	wNew = ((wOld * autoCorr[b]) + grad) / (autoCorr[b]);
                	changed = true;
                } 
                else  {
                    temp = autoCorr[b] * wOld + grad;
                    if(temp > lambda) wNew = (temp - lambda) / autoCorr[b];
                    else 
                        if(temp < -lambda) wNew = (temp + lambda) / autoCorr[b];
                    	else wNew = 0;
                    //if (Math.abs(wNew) < killThreshold) wNew = 0; // this line is new
                    changed = (wNew != 0 || wOld != 0);
                    if (!changed) nonZero += (wNew == 0 ? 0 : 1) - (wOld == 0 ? 0 : 1);
                }
                
                if (changed) {
                    tempArray = xwTranspose[mm];
                    tempArray2 = xTranspose[b];
                    temp = wNew - wOld;
                    for (i = 0; i < n; ++i) {
                        tempArray[i] = tempArray[i] + tempArray2[i] * temp;
                    }
                    for(i = 0; i < n; ++i)	{
                        if(PRECOMPUTE_EXP) {
                            temp = tempArray[i];
                            if(temp == 0) temp2 = 1;
                            else 	if(temp > -20  && temp < 0) temp2 = table1[(int) ((temp +20) * 100)];
                        			else 	if(temp > 0 && temp < 10) temp2 = table2[(int) (temp * 1000.0)];
                        					else temp2 = Math.exp(temp);
                            eNewM[i] = temp2;
                        } else
                            eNewM[i] = Math.exp(tempArray[i]);
                    }
                    tempArray2 = Etranspose[mm];
                    for(i = 0; i < n; ++i)	S[i] += eNewM[i] - tempArray2[i];
                    
                    System.arraycopy(eNewM, 0, Etranspose[mm], 0, n);
                    w[index] = wNew;
                }
            }
            mm++;
            if (mm % m == 0) {
                mm = 0;
                b++;
                if (b % d == 0) {
                    b = 0;
                    l++;
                    if (l >= maxIterations) converged = true;
                    else {
                        numer = 0;
                        denom = 0;
                        for(i = 0; i < w.length; ++i) {
                            temp = wPrev[i] - w[i];
                            numer += temp * temp;
                            temp = wPrev[i];
                            denom += temp * temp;
                        }
                        incr = Math.sqrt(numer / denom);
                        System.arraycopy(w, 0, wPrev, 0, dTimesm);
                        
                        converged = incr < convergenceTol;
                        testZeroBasis *= nonZero / dTimesmDouble;
                    }
                }
            }
        }
        IMatrix wMatrix = MatrixFactory.makeMatrix(d, m);
        for(i = 0; i < d; ++i)
            for(int j = 0; j < m; ++j)
                wMatrix.set(w[i * m + j], i, j);
        
        LearnResult lr = new LearnResult();
        lr.setIterations(l);
        lr.setWeights(wMatrix);
        return lr;
    }
}
