package edu.duke.cs.smlr.learn.learners;


import java.util.ArrayList;

import edu.duke.cs.smlr.factories.MatrixFactory;
import edu.duke.cs.smlr.learn.classifiers.DefaultClassifier;
import edu.duke.cs.smlr.learn.interfaces.IClassifier;
import edu.duke.cs.smlr.learn.interfaces.ILearner;
import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.learn.results.LearnResult;
import edu.duke.cs.smlr.util.Constants;

/**
 * Uses Iterative Optimization to learn weights
 * 
 * @author Jason Bosko, Matt Edwards
 */

public class IterativeLearner extends AbstractLearner implements ILearner {

	public LearnResult learnWeights(IMatrix X, IMatrix Y, double lambda,
			double convergenceTol, double killThreshold, IMatrix startAt,
			int method, boolean overrelaxed, boolean useless, int maxIterations) {

		/** **************** */
		/** Preliminaries * */
		/** **************** */
	    
	    Y = processClasses(Y);
	    
	    int N = Y.numRows();
		int M = Y.numCols();
		int d = X.numCols();
		// System.out.println(Y);

		/** ********************************** */
		/** Precomputing and Initialization * */
		/** ********************************** */

		int bigSize = (M - 1) * d;

		// Finding B
		IMatrix B;
		IMatrix temp = MatrixFactory.makeMatrix(M - 1, 1).add(0.5).diag();
		temp.add(MatrixFactory.makeMatrix(M - 1, M - 1).add(-1).mult(0.5 / (double) M));
		B = temp.kronecker(X.copy().transpose().mult(X));

		IMatrix Binv = MatrixFactory.makeMatrix(0, 0), BinvB = null;
		if (method == Constants.PRIOR_GAUSSIAN) {
			// Finding (B + lambda*I) inverse and that times B
			Binv = MatrixFactory.makeMatrix(bigSize, 1).add(lambda).diag().add(B).inverse();
			BinvB = Binv.mult(B);
		}
		if (method == Constants.PRIOR_NONE) {
			IMatrix XTransposeXInv = X.copy().transpose().mult(X).inverse();
			Binv = MatrixFactory.makeMatrix(M - 1, 1).add(2).diag().add(1).kronecker(
					XTransposeXInv);
		}

		// Determining starting weights
		IMatrix w = MatrixFactory.makeMatrix(bigSize, 1);
		if (startAt == null) {
			IMatrix prob = Y.copy();
			prob.add(1 / (double) M).mult(0.5);
			temp = Y.chunk(0, 0, N, M - 1).add(
					prob.chunk(0, 0, N, M - 1).mult(-1));
			IMatrix grad = X.copy().transpose().mult(temp).reshape(bigSize, 1);
			double epsilon = 1e-6;
			IMatrix Bcopy = B.copy();
			Bcopy.add(MatrixFactory.makeMatrix(bigSize, 1).add(epsilon).diag());
			w = Bcopy.inverse().mult(grad);
		} else
			w = startAt;

		// Initializing
		boolean converged = false;
		int iter = 0;
		double eta = 1;
		double etaFactor = 1.1;

		/** ************************* */
		/** Iterative Optimization * */
		/** ************************* */

		IClassifier myClassifier = new DefaultClassifier();

		while (!converged) {
			// Pruning
			ArrayList keep = new ArrayList();
			for (int k = 0; k < bigSize; ++k)
				if (Math.abs(w.get(k, 0)) > killThreshold)
					keep.add(new Integer(k));
				else
					w.set(0, k, 0);
			int wSize = keep.size();
			if (wSize == 0)
				break;
			ArrayList zeroList = new ArrayList();
			zeroList.add(new Integer(0));

			// Finding gamma, probabilities, and gradient
			IMatrix gamma = MatrixFactory.makeMatrix(wSize, 1);
			for (int k = 0; k < wSize; ++k) {
				int index = ((Integer) keep.get(k)).intValue();
				gamma.set(Math.sqrt(Math.abs(w.get(index, 0))), k, 0);
			}
			IMatrix probabilities = myClassifier.probabilities(X, w);
			temp = Y.chunk(0, 0, N, M - 1).add(
					probabilities.chunk(0, 0, N, M - 1).mult(-1));
			IMatrix gradient = X.copy().transpose().mult(temp).reshape(bigSize, 1);

			IMatrix w_new_keep = null, w_new;

			// Specific to Laplacian (l-1 penalty) prior
			if (method == Constants.PRIOR_LAPLACIAN) {
				IMatrix gammaCopy = gamma.copy();
				temp = MatrixFactory.makeMatrix(wSize, 1).add(lambda).diag();
				temp.add(B.chunk(keep, keep).termMult(
						gamma.mult(gamma.copy().transpose())));
				temp.inverse();
				w_new_keep = temp.mult(
						gamma.termMult(B.chunk(keep, keep).mult(
								w.chunk(keep, zeroList)).add(
								gradient.chunk(keep, zeroList)))).termMult(
						gammaCopy);
			}

			// Specific to Gaussian (l-2 penalty) prior
			if (method == Constants.PRIOR_GAUSSIAN)
				w_new_keep = BinvB.chunk(keep, keep).mult(
						w.chunk(keep, zeroList)).add(
						Binv.chunk(keep, keep).mult(
								gradient.chunk(keep, zeroList)));

			// Specific to Maximum Likelihood
			if (method == Constants.PRIOR_NONE)
				w_new_keep = Binv.chunk(keep, keep).mult(
						gradient.chunk(keep, zeroList)).add(
						w.chunk(keep, zeroList));

			w_new = w_new_keep.expand(keep, bigSize);

			// Overrelaxed Optimization
			if (overrelaxed) {
				IMatrix P = myClassifier.probabilities(X, w_new);
				double logLike = P.log().termMult(Y).sum();
				double logPost = logLike - lambda * w_new.sumAbs();
				IMatrix w_over = w.copy()
						.add(
								w_new.copy().add(w.copy().mult(-1))
										.mult(eta));
				P = myClassifier.probabilities(X, w_over);
				double logLikeOver = P.log().termMult(Y).sum();
				double logPostOver = logLikeOver - lambda * w_over.sumAbs();
				if ((logPostOver - logPost) < 0)
					eta = 1;
				else {
					eta *= etaFactor;
					w_new = w_over;
					logPost = logPostOver;
				}
			}

			// Check for convergence
			IMatrix wCopy = w.copy(), w_newCopy = w_new.copy();
			double nor = w_newCopy.add(wCopy.mult(-1)).norm() / w.norm();
			converged = ((nor < convergenceTol) || (Double.isNaN(nor)));
			w = w_new.copy();

			iter++;
			if (iter >= maxIterations)
				converged = true;

		}

		LearnResult results = new LearnResult();
		results.setWeights(w.reshape(d, M-1));
		results.setIterations(iter);
		return results;
	}
}
