package edu.duke.cs.smlr.learn.properties;

import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.util.Constants;

/**
 * Class that encapsulates all the settings needed to perform a single learning
 * task
 * 
 * @author Matt Edwards, Jason Bosko
 */
public class LearnProperty {

	private IMatrix data;
	private IMatrix classifications;
	private String fileName;
	
	private double lambda;
	private double convTol;
	private double killThreshold;
	private Object kernelPar;
	private int maxIters;
	private int kernel;
	private int prior;
	private int method; // algorithm
	
	private boolean useOverrelaxation;
	private boolean cacheExp;
	private boolean cacheKer;
	
	private double bias;
	private boolean normalize;

	private boolean display = true;

	public LearnProperty() {
		display = true;
	}
	
	/* Creates a hopefully fairly unique hashcode based on multiple-choice values
	 * in the GUI
	 * (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
	    return (int) (
	    	lambda * 1000.0 +
	    	convTol * 1000.0 +
	    	killThreshold * 1000.0 +
	    	(kernelPar != null ? kernelPar.hashCode() : 0) +
	    	maxIters +
	    	kernel * kernel * kernel % 1729 +
	    	prior * prior * prior % 6781+
	    	method * method * method % 1233 +
	    	(bias+10) * (bias+20) * (bias+30) % 973 +
	    	(normalize ? 223 : 0)
	    	);	    
	}

	public LearnProperty(LearnProperty l) {
		data = l.getTrainingData(); // TODO: this has bad potential
		classifications = l.getClassifications(); // todo: this too
		fileName = l.getFileName();

		lambda = l.getLambda();
		convTol = l.getConvTol();
		killThreshold = l.getKillThreshold();
		kernelPar = l.getKernelPar();

		maxIters = l.getMaxIters();
		kernel = l.getKernel();

		prior = l.getPrior();
		method = l.getMethod();

		useOverrelaxation = l.isUseOverrelaxation();
		cacheExp = l.isCacheExp();
		cacheKer = l.isCacheKer();
		
		bias = l.getBias();
		normalize = l.isNormalize();
	}

	public String toString() {
		if (!display)
			return "";
		String s = new String();
		final String sep = "\n";
		s += "File name: " + fileName;
		s += sep + "Lambda = " + lambda;
		s += sep + "Convergence tolerance = " + convTol;
		s += sep + "Kill threshold = " + killThreshold;
		if (kernelPar != null && kernel != Constants.KERNEL_DIRECT)
			s += sep + "Kernel parameter = " + kernelPar.toString();
		s += sep + "Max iterations = " + maxIters;
		s += sep + "Kernel: " + Constants.toString(kernel);
		s += sep + "Prior: " + Constants.toString(prior);
		s += sep + "Method: " + Constants.toString(method);
		if (useOverrelaxation)
			s += sep + "Using Overrelaxation";
		if (cacheExp)
		    s += sep + "Caching exponentials";
		if (cacheKer & kernel != Constants.KERNEL_DIRECT)
			s += sep + "Caching kernel functions";
		if (bias != 0) s += sep + "Added bias term of "+bias;
		if (normalize)
			s += sep + "Normalizing data";
		else
			s += sep + "Not normalizing data";

		return s;
	}
	
	public String toQuickString() {
		String kparstr = "";
		if (kernel == Constants.KERNEL_POLYNOMIAL || kernel == Constants.KERNEL_RBF) kparstr = ", Kernel Paramter = "+kernelPar.toString();
		return  "Lambda = "+lambda+
				", Kernel = "+Constants.toString(kernel)+
				kparstr+
				", Prior = "+Constants.toString(prior)+
				", Method: "+Constants.toString(method)+
				", Bias: "+bias+
				", Normalize: "+normalize;
	}

//	public String changingInformation() {
//		String s = new String();
//
//		s += "Lambda: " + lambda;
//		s += ", Kernel: ";
//		switch (kernel) {
//		case PropertyConstants.KERNEL_COSINE:
//			s += "Cosine";
//			break;
//		case PropertyConstants.KERNEL_DIRECT:
//			s += "Direct";
//			break;
//		case PropertyConstants.KERNEL_LINEAR:
//			s += "Linear";
//			break;
//		case PropertyConstants.KERNEL_POLYNOMIAL:
//			s += "Polynomial, param = "+((Double)kernelPar).doubleValue();
//			break;
//		case PropertyConstants.KERNEL_RBF:
//			s += "RBF, param = "+((Double)kernelPar).doubleValue();
//			break;
//		default:
//			throw new IllegalArgumentException("Unrecognized kernel");
//		}
//		s += ", Algorithm: ";
//		switch (method) {
//		case PropertyConstants.ALGORITHM_ITERATIVE:
//			s += "Iterative";
//			break;
//		case PropertyConstants.ALGORITHM_STEPWISE:
//			s += "Stepwise";
//			break;
//		default:
//			throw new IllegalArgumentException("Unrecognized method");
//		}
//		s += ", Prior: ";
//		switch (prior) {
//		case PropertyConstants.PRIOR_LAPLACIAN:
//			s += "Laplacian";
//			break;
//		case PropertyConstants.PRIOR_GAUSSIAN:
//			s += "Gaussian";
//			break;
//		case PropertyConstants.PRIOR_ML:
//			s += "ML";
//			break;
//		default:
//			throw new IllegalArgumentException("Unrecognized method");
//		}
//		
//		s += ", Add Bias: "+addBias;
//		return s;
//	}

	/**
	 * @return Returns the display.
	 */
	public boolean isDisplay() {
		return display;
	}

	/**
	 * @param display
	 *            The display to set.
	 */
	public void setDisplay(boolean display) {
		this.display = display;
	}

	/**
	 * @return Returns the classifications.
	 */
	public IMatrix getClassifications() {
		return classifications;
	}

	/**
	 * @return Returns the convTol.
	 */
	public double getConvTol() {
		return convTol;
	}

	/**
	 * @return Returns the data.
	 */
	public IMatrix getTrainingData() {
		return data;
	}

	/**
	 * @return Returns the kernel.
	 */
	public int getKernel() {
		return kernel;
	}

	/**
	 * @return Returns the kernelPar.
	 */
	public Object getKernelPar() {
		return kernelPar;
	}

	/**
	 * @return Returns the lambda.
	 */
	public double getLambda() {
		return lambda;
	}

	/**
	 * @return Returns the maxIters.
	 */
	public int getMaxIters() {
		return maxIters;
	}

	/**
	 * @return Returns the method.
	 */
	public int getMethod() {
		return method;
	}

	/**
	 * @return Returns the prior.
	 */
	public int getPrior() {
		return prior;
	}

	/**
	 * @param classifications
	 *            The classifications to set.
	 */
	public void setClassifications(IMatrix classifications) {
		this.classifications = classifications;
	}

	/**
	 * @param convTol
	 *            The convTol to set.
	 */
	public void setConvTol(double convTol) {
		this.convTol = convTol;
	}

	/**
	 * @param data
	 *            The data to set.
	 */
	public void setTrainingData(IMatrix data) {
		this.data = data;
	}

	/**
	 * @param kernel
	 *            The kernel to set.
	 */
	public void setKernel(int kernel) {
		this.kernel = kernel;
	}

	/**
	 * @param kernelPar
	 *            The kernelPar to set.
	 */
	public void setKernelPar(Object kernelPar) {
		this.kernelPar = kernelPar;
	}

	/**
	 * @param lambda
	 *            The lambda to set.
	 */
	public void setLambda(double lambda) {
		this.lambda = lambda;
	}

	/**
	 * @param maxIters
	 *            The maxIters to set.
	 */
	public void setMaxIters(int maxIters) {
		this.maxIters = maxIters;
	}

	/**
	 * @param method
	 *            The method to set.
	 */
	public void setMethod(int method) {
		this.method = method;
	}

	/**
	 * @param prior
	 *            The prior to set.
	 */
	public void setPrior(int prior) {
		this.prior = prior;
	}

	/**
	 * @return Returns the useOverrelaxation.
	 */
	public boolean isUseOverrelaxation() {
		return useOverrelaxation;
	}

	/**
	 * @param useOverrelaxation
	 *            The useOverrelaxation to set.
	 */
	public void setUseOverrelaxation(boolean useOverrelaxation) {
		this.useOverrelaxation = useOverrelaxation;
	}

	/**
	 * @return Returns the killThreshold.
	 */
	public double getKillThreshold() {
		return killThreshold;
	}

	/**
	 * @param killThreshold
	 *            The killThreshold to set.
	 */
	public void setKillThreshold(double killThreshold) {
		this.killThreshold = killThreshold;
	}

	/**
	 * @return Returns the addBias.
	 */
	public double getBias() {
		return bias;
	}

	/**
	 * @param addBias
	 *            The addBias to set.
	 */
	public void setBias(double bias) {
		this.bias = bias;
	}

	/**
	 * @return Returns the fileName.
	 */
	public String getFileName() {
		return fileName;
	}

	/**
	 * @param fileName
	 *            The fileName to set.
	 */
	public void setFileName(String fileName) {
		this.fileName = fileName;
	}

	public boolean isNormalize() {
		return normalize;
	}

	public void setNormalize(boolean normalize) {
		this.normalize = normalize;
	}
	
    /**
     * @return Returns the cacheExp.
     */
    public boolean isCacheExp() {
        return cacheExp;
    }
    /**
     * @param cacheExp The cacheExp to set.
     */
    public void setCacheExp(boolean cacheExp) {
        this.cacheExp = cacheExp;
    }

	public boolean isCacheKer() {
		return cacheKer;
	}

	public void setCacheKer(boolean cacheKer) {
		this.cacheKer = cacheKer;
	}	
}
