package edu.duke.cs.smlr.learn.results;

/**
 * Contains the results of finding the ROC data.
 * 
 * @author Jason Bosko
 */
public class ROCResults {
	private double thresh[];
	private int truePos[], falsePos[];
	private double TPR[], FPR[];
	private int totalPos, totalNeg;
	
	private int bestRateIndex;
	private int bestCountIndex;
	
	private double optimumRatioThresh;
	private int optimumRatioTruePos;
	private int optimumRatioFalsePos;
	private double optimumRatioTPR;
	private double optimumRatioFPR;
	
	/**
	 * Constructs an ROCResults. All arrays are parallel in elements with the same
	 * index are associated with each other.
	 * @param thr The thresholds used to generate the ROC data
	 * @param tp The count of true positives for each threshold
	 * @param fp The count of false positives for each threshold
	 * @param tpr The rate of true positives for each threshold
	 * @param fpr The rate of false positives for each threshold
	 * @param totalPos The total number of positive samples in the test
	 * @param totalNeg The total number of negative samples in the test
	 */
	public ROCResults(double thr[], int tp[], int fp[],
					  double tpr[], double fpr[],
					  int totalPos, int totalNeg) {
		thresh = thr;
		truePos = tp; falsePos = fp;
		TPR = tpr; FPR = fpr;
		this.totalPos = totalPos; this.totalNeg = totalNeg;
		findBestRate();
		findBestCount();
	}
	
	/**
	 * Sets the data needed to report the true/false positive data
	 * for the threshold defined by the ratio of positives to the data
	 * @param th The threshold defined by the ratio of positives to the data
	 * @param tp The count of true positives at the optimum threshold
	 * @param fp The count of false positives at the optimum threshold
	 */
	public void setOptimumRatio(double th, int tp, int fp) {
		optimumRatioThresh = th;
		optimumRatioTruePos = tp;
		optimumRatioFalsePos = fp;
		optimumRatioTPR = (double) tp / (double) totalPos;
		optimumRatioFPR = (double) fp / (double) totalNeg;
	}
	
	/**
	 * Prints the results obtained by performing ROC analysis.
	 * @param pw The PrintWriter to print the data to
	 */
	public String printResults() {
		int total = thresh.length;
		String s = "";
		for (int i = 0 ; i < total; i++) {
			s+= "Threshold = "+thresh[i]+" -->\n";
			s += "\t True Positives  = "+truePos[i] +", True Positive Rate  = "+TPR[i]+"\n";
			s += "\t False Positives = "+falsePos[i]+", False Positive Rate = "+FPR[i]+"\n";
		}
		s += "\n";
		s += "The area underneath the ROC curve is: "+computeAreaUnderROC()+"\n";
		s += "\n";
		s += "The best threshold with respect to rates is: "+thresh[bestRateIndex]+" -->"+"\n";
		s += "\t True count: "+truePos[bestRateIndex]+", True rate: "+TPR[bestRateIndex]+"\n";
		s += "\t False count: "+falsePos[bestRateIndex]+", False rate: "+FPR[bestRateIndex]+"\n";
		
		s += "The best threshold with respect to counts is: "+thresh[bestCountIndex]+" -->"+"\n";
		s += "\t True count: "+truePos[bestCountIndex]+", True Rate: "+TPR[bestCountIndex]+"\n";
		s += "\t False count: "+falsePos[bestCountIndex]+", False rate: "+FPR[bestCountIndex]+"\n";

		s += "The threshold with respect to the ratio of positives to negatives is: "+optimumRatioThresh+" -->"+"\n";
		s += "\t True count: "+optimumRatioTruePos+", True rate:"+optimumRatioTPR+"\n";
		s += "\t False count: "+optimumRatioFalsePos+", False rate: "+optimumRatioFPR;
		return s;
	}
	
//	/**
//	 * Prints the results obtained by performing ROC analysis.
//	 * @param pw The PrintWriter to print the data to
//	 */
//	public void printResults(PrintWriter pw) {
//		int total = thresh.length;
//		pw.println("\n==========ROC statistics==========");
//		for (int i = 0 ; i < total; i++) {
//			pw.println("Threshold = "+thresh[i]+" -->");
//			pw.println("\t True Positives  = "+truePos[i] +", True Positive Rate  = "+TPR[i]);
//			pw.println("\t False Positives = "+falsePos[i]+", False Positive Rate = "+FPR[i]);
//		}
//		pw.println();
//		pw.println("The area underneath the ROC curve is: "+computeAreaUnderROC());
//		pw.println();
//		pw.println("The best threshold with respect to rates is: "+thresh[bestRateIndex]+" -->");
//		pw.println("\t True count: "+truePos[bestRateIndex]+", True rate: "+TPR[bestRateIndex]);
//		pw.println("\t False count: "+falsePos[bestRateIndex]+", False rate: "+FPR[bestRateIndex]);
//		
//		pw.println("The best threshold with respect to counts is: "+thresh[bestCountIndex]+" -->");
//		pw.println("\t True count: "+truePos[bestCountIndex]+", True Rate: "+TPR[bestCountIndex]);
//		pw.println("\t False count: "+falsePos[bestCountIndex]+", False rate: "+FPR[bestCountIndex]);
//
//		pw.println("The threshold with respect to the ratio of positives to negatives is: "+optimumRatioThresh+" -->");
//		pw.println("\t True count: "+optimumRatioTruePos+", True rate:"+optimumRatioTPR);
//		pw.println("\t False count: "+optimumRatioFalsePos+", False rate: "+optimumRatioFPR);
//		pw.println("==========End of ROC statistics==========");
//	}
	
	/**
	 * Finds the threshold that produces the best results with 
	 * respect to true/false positive rates.
	 */
	private void findBestRate() {
		int index = 0;
		int total = thresh.length;
		double bestRate = -1;
		for (int i = 0; i < total; i++) {
			if (bestRate < TPR[i] + (1-FPR[i])) {
				bestRate = TPR[i] + (1-FPR[i]);
				index = i;
			}
		}
		bestRateIndex = index;
	}
	
	/**
	 * Finds the threshold that produces the best results with 
	 * respect to true/false positive counts.
	 */
	private void findBestCount() {
		int index = 0;
		int total = thresh.length;
		int bestCount = -1;
		for (int i = 0; i < total; i++) {
			if (bestCount < truePos[i] + (totalNeg-falsePos[i])) {
				bestCount = truePos[i] + (totalNeg-falsePos[i]);
				index = i;
			}
		}
		bestCountIndex = index;
	}
	
	/**
	 * Computes the area underneath the ROC curve
	 * @return The area underneath the ROC curve
	 */
	private double computeAreaUnderROC() {
		double currFPR, prevFPR;
		double prevTPR;
		double AUC = 0;
		// Compute the area by summing the different rectangles formed from 
		// the true and false positive rates.
		for (int i = 0; i < FPR.length-1; i++) {
			currFPR = FPR[i]; prevFPR = FPR[i+1]; prevTPR = TPR[i+1];
			AUC += prevTPR * (currFPR - prevFPR);
		}
		return AUC;
	}
}
