package edu.duke.cs.smlr.util;


import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.TreeSet;

import edu.duke.cs.smlr.factories.MatrixFactory;
import edu.duke.cs.smlr.learn.interfaces.IMatrix;

/**
 * A general-purpose reader to read input data without knowing much about the
 * format. It assumes that the data file is in row format with features followed
 * by classification or vice-versa. It guesses that the column with fewer
 * distinct values is the classification column.
 * 
 * @author Matt Edwards, Jason Bosko
 */
public class AdaptiveReader {
	private BufferedReader reader;
	private IMatrix X;
	private IMatrix Y;
	private ClassificationMap classificationMap;
	private int M;
	private int classesColumn;

	/**
	 * A specialized read for the adaptive reader
	 * 
	 * @param filename
	 *            the name of the file to read from
	 * @param classesCol
	 *            the column where classifications are, using READER* constants
	 *            from PropertyConstants class
	 * @throws FileNotFoundException
	 * 			  if... the file can't be found
	 * @throws HaltingException
	 * 			  if the data is invalid 
	 */
	public void read(String filename, int classesCol) throws FileNotFoundException, HaltingException {
		this.classesColumn = classesCol;
		read(filename);
	}

	private void read(String filename) throws FileNotFoundException, HaltingException {
		reader = new BufferedReader(new FileReader(filename), 8192 * 16);

		int n = 0, d = -1;
		String line;

		ArrayList matrixRows = new ArrayList(); //will store String[]
		int lines = 0;
		while (true) {
			lines++;
			try {
				if (!reader.ready())
					break;
				line = reader.readLine();
			} catch (IOException e) {
				//System.out.println("BREAK");
				break;
			}
			if (line == null)
				continue;
			String tokens[] = line.split("[ ,:;	]");
			LinkedList ll = new LinkedList();
			ll.addAll(Arrays.asList(tokens));
			ListIterator i = ll.listIterator();
			while (i.hasNext()) {
				String curr = (String) i.next();
				if (curr.equals("") || curr.equals(" ")) {
					i.remove();
				}
			}
			tokens = new String[ll.size()];
			tokens = (String[]) ll.toArray(new String[0]);
			if (tokens.length == 0)
				continue;
			n++;
			matrixRows.add(tokens);
			if (d != -1) {
				if (d != tokens.length - 1) {
					if (matrixRows.size() == 2) {
						// maybe the first row is the only exception... then set
						// the second row as valid
						// if the data set is just plain bad, it should still
						// fail eventually
						matrixRows.remove(0);
						d = tokens.length - 1;
						n--; // decrease observation count
					} else
						throw new HaltingException("Invalid data");
				}
			} else
				d = tokens.length - 1;
		}
		
		//Now we have MatrixRows: an ArrayList of String[] that contains
		//what has been read in from the file. We need to choose which
		//column has the classifications and then parse all Strings as
		//doubles
		String currentLine[];
		
		//this figures out what column the classifications are on
		int useColumn = 0;
		if (classesColumn == Constants.READER_FIRST_COLUMN)
			useColumn = 0;
		else if (classesColumn == Constants.READER_LAST_COLUMN)
			useColumn = d;
		else if (classesColumn == Constants.READER_GUESS) {
			int firstUniqueCount = 0, lastUniqueCount = 0;
			HashSet firstCol = new HashSet();
			HashSet lastCol = new HashSet();
			for (int i = 0; i < matrixRows.size(); ++i) {
				currentLine = (String[]) matrixRows.get(i);
				String s = currentLine[0];
				if (!firstCol.contains(s)) {
					firstUniqueCount++;
					firstCol.add(s);
				}
				s = currentLine[currentLine.length-1];
				if (!lastCol.contains(s)) {
					lastUniqueCount++;
					lastCol.add(s);
				}
			}
			if (firstUniqueCount <= 1) firstUniqueCount = Integer.MAX_VALUE; //single class is not good
			if (lastUniqueCount <= 1) lastUniqueCount = Integer.MAX_VALUE; //single class is not good
			useColumn = (firstUniqueCount <= lastUniqueCount) ? 0 : d;
		}
		else if (classesColumn == Constants.READER_UNCLASSIFIED) {
			useColumn = -1;
			d++;
		}
		else
			useColumn = classesColumn;
		
		//now split up and process matrixRows
		Y = MatrixFactory.makeMatrix(n, 1);
		String[] classNames = new String[n];
		X = MatrixFactory.makeMatrix(n, d); 
		int xIndex = 0;
		double currentNumber;
		String currentString;
		for (int i = 0; i < matrixRows.size(); ++i) {
			currentLine = (String[]) matrixRows.get(i);
			xIndex = 0;
			for (int j = 0; j < currentLine.length; ++j) {
				currentString = currentLine[j];
				try {
					currentNumber = Double.parseDouble(currentString);
				}
				catch (NumberFormatException e) {
					currentNumber = (double) currentString.hashCode() % 100;
				}
				if (j == useColumn) {
					Y.set(currentNumber, i, 0);
					classNames[i] = currentString;
				}
				else {
					X.set(currentNumber, i, xIndex);
					xIndex++;
				}
			}
		}
		
		// post-process so that the classes are 1, 2, 3, ...
		// also stores the mapping of the new values (1, 2, 3, ...) to the old names
		classificationMap = new ClassificationMap();
		if (classesColumn != Constants.READER_UNCLASSIFIED) {
			TreeSet nameSet = new TreeSet();
			for (int i = 0; i < Y.numRows(); ++i) 
				if (!nameSet.contains(classNames[i])) {
					nameSet.add(classNames[i]);
				}
			ArrayList nameList = new ArrayList();
			nameList.addAll(nameSet);
			for (int i = 0; i < nameList.size(); ++i) {
				classificationMap.put(nameList.get(i), new Double(i+1));
			}
			for (int i = 0; i < Y.numRows(); ++i) {
				Y.set(nameList.indexOf(classNames[i]) + 1., i, 0);
			}
			M = (int) Y.maxAbsRowSum();
		}
		else
			M = 0;
	}

	/**
	 * Provides a deep copy of the matrix containing all the features of the data.
	 * @return A deep copy of the data matrix
	 */
	public IMatrix getX() {
		return X.copy();
	}

	
	/**
	 * Provides a deep copy of the matrix containing all the features of the data.
	 * @return A deep copy of the data matrix
	 */
	public IMatrix getY() {
		return Y.copy();
	}
	
	/**
	 * Provides the map of what classes are called in the matrix (1, 2,..., m) to what they are called in the actual data (-1, +1, yes, no, etc.)
	 * @return The classification map.
	 */
	public ClassificationMap getClassificationMap() {
		return classificationMap;
	}

	/**
	 * Provides the number of examples used in the data.
	 * @return The number of examples used in the data.
	 */
	public int getN() {
		return X.numRows();
	}

	/**
	 * Provides the number of features in the data.
	 * @return The number of features in the data.
	 */
	public int getD() {
		return X.numCols();
	}

	/**
	 * Provides the number of classes in the data.
	 * @return The number of classes in the data.
	 */
	public int getM() {
		return M;
	}
	
}
