package edu.duke.cs.smlr.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.swing.JOptionPane;

import edu.duke.cs.smlr.Graphical;
import edu.duke.cs.smlr.factories.MatrixFactory;
import edu.duke.cs.smlr.learn.KFolder;
import edu.duke.cs.smlr.learn.KernelSource;
import edu.duke.cs.smlr.learn.Learner;
import edu.duke.cs.smlr.learn.Modifier;
import edu.duke.cs.smlr.learn.Tester;
import edu.duke.cs.smlr.learn.classifiers.ClassifierThresholdFinder;
import edu.duke.cs.smlr.learn.interfaces.IMatrix;
import edu.duke.cs.smlr.learn.properties.ComboProperty;
import edu.duke.cs.smlr.learn.properties.KFoldProperty;
import edu.duke.cs.smlr.learn.properties.LearnProperty;
import edu.duke.cs.smlr.learn.properties.ModifyProperty;
import edu.duke.cs.smlr.learn.properties.TestProperty;
import edu.duke.cs.smlr.learn.results.KFoldResult;
import edu.duke.cs.smlr.learn.results.LearnResult;
import edu.duke.cs.smlr.learn.results.ROCResults;
import edu.duke.cs.smlr.learn.results.TestResult;


/**A class that contains the common backend methods for the text and graphical interfaces. 
 * @author Matt Edwards, Jason Bosko
 */
public class Core {    
    
    /**This starts the specified learn task on a background thread.  <br>
     * This should only be called from the GUI.  It will show a pop-up box when
     * done.
     * @param settings learning parameters
     */
    public static void doSettings(final String[] settings, final Graphical parent) throws ArgumentFormatException, HaltingException{
    	SwingWorker worker = new SwingWorker() {
        	public Object construct() {
        		try {
        			Core.doSettings(settings , null, null);
        		} catch (ArgumentFormatException e) {
        			e.printStackTrace();
        			JOptionPane.showMessageDialog(parent, 
        					"Error: File Not Found \n" +
        					"Please recheck all files. ", "File Not Found", JOptionPane.ERROR_MESSAGE);
        			return new Boolean(false);
        		} catch (HaltingException e) {
        			JOptionPane.showMessageDialog(parent, "Error: Fatal Error Occured. Restart the test. " + (e.getMessage() != null ? "\nThe error reported is: "+e.getMessage() : ""), "Fatal Error", JOptionPane.ERROR_MESSAGE);
        			return new Boolean(false);
        		} catch (OutOfMemoryError e) {
        			JOptionPane.showMessageDialog(parent, 
            				Messages.ERROR_MEMORY,
							"Error", JOptionPane.ERROR_MESSAGE);        			
        			return new Boolean(false);
        		}
        		return new Boolean(true);
        	}
        	public void finished() {
        		if (getValue().equals(new Boolean(true))) {
        		JOptionPane.showMessageDialog(parent,
        				"Learning task completed.",
						"Finished", JOptionPane.INFORMATION_MESSAGE); }
        		else {   			
        		}
        		parent.reenableLearn();
        	}
    	};
        worker.start();
    }

    /**Retrieves a filename from an arrayList as follows (braces are merely notational): <br>
     * ArrayList list = { [hello], ["file ], [name.txt"], [etc] } <br>
     * parseFileName(list, 1) = [file name.txt]
     * 
     * @param options the arrayList of strings
     * @param index the index to start reading from
     * @return the parsed filename
     */
    public static String parseFileName(ArrayList options, int index) {
		String fileName = "";
		String next = (String) options.get(index);
		if (next.startsWith("\"") && next.endsWith("\""))
			fileName = next.substring(1, next.length() - 1);
		else if (next.startsWith("\"")) {
			fileName += next.substring(1);
			index++;
			next = (String) options.get(index);
			while (!next.endsWith("\"")) {
				fileName += " " + next;
				index++;
				next = (String) options.get(index);
			}	
			fileName +=	" " + next.substring(0, next.length()-1);
		}
		else 
			fileName = next;
		return fileName;		        
    }
    
	/**Takes a set of properties and computes the desired learning task
	 * @param columnChoice the column choice constant for the adaptiveReader
	 * @throws ArgumentFormatException if it encounters file trouble
	 */
	public static void process(ModifyProperty modifyProp[],
			LearnProperty learnProp, TestProperty testProp,
			KFoldProperty kFoldProp, int testOption,
			OutputController output, int columnChoice,
			String outFile, String settings, String id) throws ArgumentFormatException, HaltingException {
		double startTime = System.currentTimeMillis();
		
		// Need to make new directories (if needed)
		File f = new File(outFile);
		try {
			if (!f.exists()) f.mkdirs();
		} catch (SecurityException e) {
			throw new HaltingException("Cannot make directory. Check permissions for the directory name.");
		}
		
		char ending = '/';
		if (outFile.indexOf("\\") != -1) ending = '\\';
		if (!outFile.endsWith("/") && !outFile.endsWith("\\")) outFile += ending;
		output.setFiles(outFile+"smlr.log.test"+id+".txt", outFile+"smlr.summary.test"+id+".txt");
		
		output.printManual(OutputController.BOTH, "Command used to run task: " + settings + "\n\n");
		output.printManual(OutputController.BOTH, "Learning task started.");
		output.print(OutputController.SUMMARY, OutputController.BEHAVIOR, "Overall test pattern: " + Constants.toString(testOption));
		
		//****************************************************************
		// Set up the ComboProperty array that dictates the test.
		// If testOption is testontesting or unlabeled, there
		// will only be one element in this array (because there will only
		// be one test). If testOption is kfoldtest or loocv, then there 
		// will be multiple elements in the array.
		//****************************************************************
		
		AdaptiveReader myReader = new AdaptiveReader();
		ComboProperty comboProperty[] = new ComboProperty[1];
		
		// Every type of test needs training data
		try {		     
			if (learnProp.getFileName() == null) throw new FileNotFoundException("fake"); // necessary?
			myReader.read(learnProp.getFileName(), columnChoice);
		} catch (FileNotFoundException e) {
			throw new ArgumentFormatException("Training File Not Found");
		} catch (HaltingException e) {
			throw new HaltingException("Invalid Training File Given");
		}
		ClassificationMap classMap = myReader.getClassificationMap();
		
		learnProp.setTrainingData(Modifier.modifyData(myReader.getX(),	modifyProp[0]));
		learnProp.setClassifications(Modifier.modifyClasses(myReader.getY(), modifyProp[0]));
		
		// Find out how many classes there are
		HashSet hs = new HashSet();
		for (int i = 0; i < learnProp.getClassifications().numRows(); i++) hs.add(new Double(learnProp.getClassifications().get(i, 0)));
		int numclasses = hs.size();
		
		// Important to know if this is a binary or multiclass classifier
		boolean binary = false;
		if (numclasses == 2) binary = true;
		
		// These are used if a binary classifier is learned (used for ROC)
		IMatrix finalProbs = MatrixFactory.makeMatrix(learnProp.getTrainingData().numRows(), 2);
		IMatrix finalClasses = learnProp.getClassifications();
		
		// Useful to know how many examples there are for each class label
		int dataPoints = learnProp.getTrainingData().numRows();
		int numInClass[] = new int[numclasses];
		for (int i = 0; i < numclasses; i++)
			numInClass[i] = 0;
		for (int i = 0; i < learnProp.getClassifications().numRows(); i++) {
			numInClass[((int) learnProp.getClassifications().get(i,0))-1]++;
		}
		
		// This option tests a classifier on a separate file of data
		if (testOption == Constants.TEST_TESTING_DATA) {
			String s = testProp.getTestingFileName();
			testProp = new TestProperty(learnProp);
			testProp.setTestingFileName(s);
			testProp.setWeightFileName(testProp.getWeightFileName());
			AdaptiveReader testReader = new AdaptiveReader();
			try {
				if (testProp.getTestingFileName() == null)
					throw new FileNotFoundException("fake"); // necessary?
				testReader.read(testProp.getTestingFileName(), columnChoice);

			} catch (FileNotFoundException e) {
				throw new ArgumentFormatException("Invalid / No Testing File Given");
			}
			testProp.setTestData(Modifier.modifyData(testReader.getX(),	modifyProp[1]));
			testProp.setTrainingData(learnProp.getTrainingData());
			if (testProp.getTestData().numCols() != testProp.getTrainingData().numCols()) {
				throw new HaltingException("Mismatched number of features: "+testProp.getTestData().numCols()+" features in test data and "+testProp.getTrainingData().numCols()+" features in training data.");
			}
			testProp.setClassifications(Modifier.modifyClasses(testReader.getY(), modifyProp[1]));
			comboProperty[0] = new ComboProperty(learnProp, testProp);
			comboProperty[0].setLearnDataMap(DataPointMap.makeSequentialMap(learnProp.getTrainingData().numRows()));
			comboProperty[0].setTestDataMap(DataPointMap.makeSequentialMap(testProp.getTestData().numRows()));
			finalProbs = MatrixFactory.makeMatrix(testProp.getTestData().numRows(), numclasses);
			finalClasses = testProp.getClassifications();
		}
		
		// This options uses k-fold cross validation to evaluate a classifier
		if (testOption == Constants.TEST_KFOLD) {
			kFoldProp.setLearnProperty(learnProp);
			comboProperty = KFolder.kfold(kFoldProp);
			output.print(OutputController.BOTH, OutputController.KFOLD_PROPERTY, kFoldProp.toString());
		}
		// This options uses leave-one-out cross validation to evaluate a classifier
		if (testOption == Constants.TEST_LOOCV) {
			kFoldProp.setLearnProperty(learnProp);
			kFoldProp.setK(learnProp.getTrainingData().numRows());
			kFoldProp.setBehavior(kFoldProp.getBehavior());
			comboProperty = KFolder.kfold(kFoldProp);
			output.print(OutputController.BOTH, OutputController.KFOLD_PROPERTY, kFoldProp.toString());
		}
		
		// This option uses the classifier to classify unlabeled data in another file. It is set up like the 'testontesting'
		// option, with the unlabeled data as the test data, but certain options will be turned off later
		if (testOption == Constants.TEST_UNLABELED) {
			String s = testProp.getTestingFileName();
			testProp = new TestProperty(learnProp);
			testProp.setTestingFileName(s);
			testProp.setWeightFileName(testProp.getWeightFileName());
			AdaptiveReader testReader = new AdaptiveReader();
			try {
				if (testProp.getTestingFileName() == null)
					throw new FileNotFoundException("fake"); // necessary?
				testReader.read(testProp.getTestingFileName(), Constants.READER_UNCLASSIFIED);

			} catch (FileNotFoundException e) {
				throw new ArgumentFormatException("Invalid / No Unlabeled File Given");
			}
			testProp.setTestData(Modifier.modifyData(testReader.getX(),	modifyProp[2]));
			testProp.setTrainingData(learnProp.getTrainingData());
			
			if (testProp.getTestData().numCols() != testProp.getTrainingData().numCols()) {
				throw new HaltingException("Mismatched number of features: "+testProp.getTestData().numCols()+" features in unlabeled data and "+testProp.getTrainingData().numCols()+" features in training data.");
			}
			
			//a classifications matrix needs to be added so that Tester.test is happy
			testProp.setClassifications(MatrixFactory.makeMatrix(testReader.getX().numRows(), 1));
			comboProperty[0] = new ComboProperty(learnProp, testProp);
			comboProperty[0].setLearnDataMap(DataPointMap.makeSequentialMap(learnProp.getTrainingData().numRows()));
			comboProperty[0].setTestDataMap(DataPointMap.makeSequentialMap(testProp.getTestData().numRows()));
			System.out.println(testProp.getTestData().numRows());
		}
		
		// This option will just learn a classifier and not evaluate it -- useful for feature selection
		if (testOption == Constants.TEST_NONE) {
			comboProperty[0] = new ComboProperty(learnProp);
			comboProperty[0].setLearnDataMap(DataPointMap.makeSequentialMap(learnProp.getTrainingData().numRows()));
		}
		
		// Output the details of the test to be run
		output.print(OutputController.BOTH, OutputController.LEARN_PROPERTY, comboProperty[0].getLearnProperty().toString());
		if (!modifyProp[0].getDeletedColumns().isEmpty() || !modifyProp[0].getDeletedRows().isEmpty()) output.print(OutputController.BOTH, OutputController.MODIFY_PROPERTY, "Modifications to training data: \n"+modifyProp[0].toString());
		if (testOption == Constants.TEST_TESTING_DATA & (!modifyProp[1].getDeletedColumns().isEmpty() || !modifyProp[1].getDeletedRows().isEmpty())) output.print(OutputController.BOTH, OutputController.MODIFY_PROPERTY, "Modifications to test data: \n"+modifyProp[1].toString());
		if (testOption == Constants.TEST_UNLABELED & (!modifyProp[2].getDeletedColumns().isEmpty() || !modifyProp[2].getDeletedRows().isEmpty())) output.print(OutputController.BOTH, OutputController.MODIFY_PROPERTY, "Modifications to unlabeled data: \n"+modifyProp[2].toString());
		if (testOption == Constants.TEST_TESTING_DATA || testOption == Constants.TEST_UNLABELED)
			output.print(OutputController.BOTH, OutputController.TEST_PROPERTY, testProp.toString());
		
		// Set up the confusion matrix
		int confu[][] = new int[numclasses][numclasses];
		
		// Useful if the type of testing is cross-validation
		KFoldResult kFoldResult = new KFoldResult();
		kFoldResult.setClassificationMap(classMap);
		int totalcorrect = 0, totalmissed = 0, totalBasisFunc = 0;
		IMatrix totalWeights = null;
		int weightCounts[][] = null;
		int totalMisclasses[] = new int[numclasses];
		for (int k=0; k < numclasses; k++) totalMisclasses[k] = 0;
		
		// Set up the caching for kernel functions -- if cache is 'null', then this is interpreted as not needing/wanting a cache
		// Kernel caching is only useful if (a) the test behavior is cross validation, and (b) a kernel is being used (not direct)
		KernelSource cache = null;		
		if ( (comboProperty[0].getLearnProperty().isCacheKer()) && 
			 (testOption == Constants.TEST_KFOLD || testOption == Constants.TEST_LOOCV) && 
			 (comboProperty[0].getLearnProperty().getKernel() != Constants.KERNEL_DIRECT) )
			cache = new KernelSource(kFoldProp.getLearnProperty());
		
		// Now its time to go through all the tests that have been set up in comboProperty
		for (int k = 0; k < comboProperty.length; ++k) {
			
			//are multiple runs being made (i.e., is it cross-validation)
			boolean multiple = (testOption == Constants.TEST_KFOLD || testOption == Constants.TEST_LOOCV);
			
			output.flush(OutputController.BOTH);
			if (multiple)
				output.printManual(OutputController.LOG, "\nTest " + (k + 1) + " of " + comboProperty.length + " ---------------------------------------------------------");
			
			// Learn the classifier for this specific test
			LearnResult lResult = Learner.learn(comboProperty[k].getLearnProperty(), comboProperty[k].getLearnDataMap(), cache);
			lResult.setClassificationMap(classMap);
			lResult.setTrainMap(comboProperty[k].getLearnDataMap());
			int choice = OutputController.SUMMARY;
			if (multiple) choice = OutputController.LOG;
			output.print(choice, OutputController.TIME_ELAPSED, "Time Elapsed: "+lResult.getTime());
			output.print(choice, OutputController.ITERATIONS, "Iterations: "+lResult.getIterations());
			output.print(choice, OutputController.BASIS_USED, "Total basis functions used: "+lResult.getBasisUsed());
			output.print(choice, OutputController.WEIGHTS, "\nWeights: \n"+lResult.getWeights());

			// Automatically test the classifier on the training data
			TestProperty trainTest = new TestProperty(comboProperty[k].getLearnProperty());
			trainTest.setWeights(lResult.getWeights());
			TestResult trainResult = Tester.test(trainTest, comboProperty[k].getLearnDataMap(), comboProperty[k].getLearnDataMap(), cache);
			trainResult.setClassificationMap(classMap);
			trainResult.setTrainMap(comboProperty[k].getLearnDataMap());
			trainResult.setTestMap(comboProperty[k].getLearnDataMap());
			output.printManual(OutputController.LOG, "\nAutomatic test of classifier on training data: \n");
			output.print(OutputController.LOG, OutputController.CLASSIFICATIONS, "Classifications: \n"+trainResult.getClassificationsString());
			output.print(OutputController.LOG, OutputController.PROBABILITIES, "Probabilities: \n"+trainResult.getProbabilitiesString());
			output.print(OutputController.LOG, OutputController.CORRECT, "Number correct: "+trainResult.getNumCorrect());
			output.print(OutputController.LOG, OutputController.INCORRECT, "Number incorrect: "+trainResult.getNumWrong());
			output.print(OutputController.LOG, OutputController.PERCENTAGE, "Percentage correct: "+(double)trainResult.getNumCorrect()*100/((double)trainResult.getNumCorrect()+(double)trainResult.getNumWrong()));
			if (trainResult.getNumWrong() > 0) output.print(OutputController.LOG, OutputController.MISCLASSIFICATIONS, trainResult.getMisclassificationsString());
			output.printManual(OutputController.LOG, "End of automatic test of classifier on training data.\n");
			
			// Evaluate the classifier using the specified method (unless the specified method is to not test the classifier)
			if (testOption != Constants.TEST_NONE) {
				
				comboProperty[k].getTestProperty().setWeights(lResult.getWeights());
				TestResult tResult = Tester.test(comboProperty[k].getTestProperty(), comboProperty[k].getTestDataMap(), comboProperty[k].getLearnDataMap(), cache);
				tResult.setTrainMap(comboProperty[k].getLearnDataMap());
				tResult.setTestMap(comboProperty[k].getTestDataMap());
				tResult.setClassificationMap(classMap);
				
				//specialized display for using the classifier on unlabeled data
				if (testOption == Constants.TEST_UNLABELED) {
					//Set it to output only the classifications and their associated probabilities
					output.setControl(OutputController.PROBABILITIES, true);
					output.setControl(OutputController.CLASSIFICATIONS, true);
					output.setControl(OutputController.CORRECT, false);
					output.setControl(OutputController.INCORRECT, false);
					output.setControl(OutputController.PERCENTAGE, false);
					output.setControl(OutputController.MISCLASSIFICATIONS, false);
					output.printManual(choice, "\nUsing classifier on unlabeled data:\n");
				}
				
				output.print(choice, OutputController.CLASSIFICATIONS, "Classifications: \n"+tResult.getClassificationsString());
				output.print(choice, OutputController.PROBABILITIES, "Probabilities: \n" +tResult.getProbabilitiesString());
				output.print(choice, OutputController.CORRECT, "Number correct: "+tResult.getNumCorrect());
				output.print(choice, OutputController.INCORRECT, "Number incorrect: "+tResult.getNumWrong());
				output.print(choice, OutputController.PERCENTAGE, "Percentage correct: "+(double)tResult.getNumCorrect()*100/((double)tResult.getNumCorrect()+(double)tResult.getNumWrong()));
				if (tResult.getNumWrong() > 0) output.print(choice, OutputController.MISCLASSIFICATIONS, tResult.getMisclassificationsString());
				
				
				// Some extra output for classifying unlabeled data
				if (testOption == Constants.TEST_UNLABELED) {
					IMatrix cl = tResult.getClassifications();
					
					ArrayList counts[] = new ArrayList[classMap.getTotalClasses()];
					for (int a = 0; a < counts.length; a++) counts[a] = new ArrayList();
					for (int a = 0; a < cl.numRows(); a++) {
						counts[(int) cl.get(a,0) - 1].add(new Integer(a+1));
					}
					//report the findings
					String s = "";
					for (int a = 0; a < classMap.getTotalClasses(); a++) {
						s+="Classified as "+classMap.getOriginalClassification(new Double(a+1))+":";
						for (Iterator it = counts[a].iterator(); it.hasNext();) {
							s+=" "+it.next();
						}
						s+="\n";
					}
					output.printManual(choice, s);
				}
				
				// Add the probabilities to the final matrix of probabilities used in calculating the ROC
				// Only do this if we are using a binary classifier and not classifying unlabeled data
				if (binary && testOption != Constants.TEST_UNLABELED) {
					for (int i = 0; i < tResult.getProbabilities().numRows(); i++) {
						int index = comboProperty[k].getTestDataMap().getOriginalDataPoint(new Integer(i));
						finalProbs.setQuick(tResult.getProbabilities().getQuick(i, 0), index, 0);
						finalProbs.setQuick(tResult.getProbabilities().getQuick(i, 1), index, 1);
					}
				}
				
				// Update the confusion matrix -- if cross-validation is not being used, this will just 
				// copy the confusion matrix from the test result (because this loop only gets run once)
				if (testOption != Constants.TEST_UNLABELED) {
					for (int a = 0; a < confu.length; a++)
						for (int b = 0; b < confu[a].length; b++)
							confu[a][b] += tResult.getConfusionMatrix()[a][b];
				}
				
				// Calculate statistics for cross-validation
				if (testOption == Constants.TEST_KFOLD || testOption == Constants.TEST_LOOCV) {	
					// If there is no kernel, then there should be as many weights as features (+1 for bias)
					// If there is a kernel, then there should be as many weights as data points in the current training set
					// Only need to create these objects on the first fold of cross-validation
					if (k == 0 && comboProperty[k].getLearnProperty().getKernel() == Constants.KERNEL_DIRECT) {
						totalWeights = MatrixFactory.makeMatrix(lResult.getWeights().numRows(), lResult.getWeights().numCols());
						weightCounts = new int[lResult.getWeights().numRows()][lResult.getWeights().numCols()];
					}
					if (k == 0 && comboProperty[k].getLearnProperty().getKernel() != Constants.KERNEL_DIRECT) {
						totalWeights = MatrixFactory.makeMatrix(dataPoints+1,lResult.getWeights().numCols());
						weightCounts = new int[dataPoints+1][lResult.getWeights().numCols()];
					}
					
					// Update the running count of misclassifications, total right, wrong, and bases
					for (Iterator it = tResult.getMisclassifications().iterator(); it.hasNext();) {
						int ind = ((Integer)it.next()).intValue();
						totalMisclasses[(int) tResult.getRealClassifications().get(ind, 0) - 1]++;
					}
					totalcorrect += tResult.getNumCorrect();
					totalmissed += tResult.getNumWrong();
					totalBasisFunc += lResult.getBasisUsed(); 
					
					// Update the running values of weights
					if (comboProperty[k].getLearnProperty().getKernel() == Constants.KERNEL_DIRECT) {
						// If there is no kernel, then weights for corresponding features will always be in the same place
						totalWeights.add(lResult.getWeights());
						for (int i = 0; i < totalWeights.numRows(); ++i) 
							for (int j = 0; j < totalWeights.numCols(); ++j) 
								if (Math.abs(lResult.getWeights().get(i, j)) > 0)
									weightCounts[i][j]++;
						kFoldResult.divFactor = kFoldProp.getK();
					}
					else {
						// If there is a kernel, then the weights are specific to kernel functions on certain
						// points. For example, feature 5 really corresponds to the kernel function with 
						// data point 5, so we need to look up what data point 5 was in the original data 
						// set and then add the weight in the right place.
						IMatrix w = lResult.getWeights();
						for (int i = 0; i < w.numRows(); i++) {
							int index = -1;
							if (i!=w.numRows()-1) index = comboProperty[k].getLearnDataMap().getOriginalDataPoint(new Integer(i));
							else index = dataPoints;
							for (int q = 0; q < w.numCols(); q++) {
								totalWeights.set(totalWeights.get(index, q)+w.get(i, q), index, q);
								if (w.get(i, q) != 0) weightCounts[index][q]++;
							}
						}
						// In k-fold or loocv, each point is left out once, so while there are 'k' folds,
						// each data point only has a chance to be in the weights 'k'-1 times.
						kFoldResult.divFactor = kFoldProp.getK() - 1;
					}
				}
			}
			
			// Create the string of features to exclude if cross-validation is not being used (so we know that the testing is finished)
			if (testOption != Constants.TEST_KFOLD && testOption != Constants.TEST_LOOCV) {
				String ret = "";
				int excluded = 0;
				for (int a = 0; a < lResult.getWeights().numRows(); a++) {
					boolean excludeMe = true;
					// Exclude a weight only if all the entries in a row are zero
					for (int b = 0; b < lResult.getWeights().numCols(); b++) {
						if (Math.abs(lResult.getWeights().get(a, b)) > 0) excludeMe = false;
					}
					if (excludeMe) {
						if (excluded != 0) ret+=", ";
						ret += ""+(a+1);
						excluded++;
					}
				}
				ret = "\nFeatures (one-based) to be excluded ("+excluded+" excluded): " + ret;
				output.printManual(choice, ret);
			}
		}
		// We are done with this cross validation, so we can clear the kernel cache
		if (cache != null) cache.clear();
		
		// Store the statistics recorded during cross-validation						
		if (testOption == Constants.TEST_KFOLD || testOption == Constants.TEST_LOOCV) {
			int ch = OutputController.SUMMARY;
			output.printManual(ch, "\n=== Overall Summary =================================================");
			kFoldResult.setNumInClass(numInClass);
			double endTime = System.currentTimeMillis();
			double totalTime = (endTime - startTime) / 1000;
			kFoldResult.setAvgBasisFunc((double)totalBasisFunc/(double)kFoldProp.getK());
			kFoldResult.setTotalTime(totalTime);
			kFoldResult.setNumOfClassWrong(totalMisclasses);
			kFoldResult.setAverageWeights(totalWeights);
			kFoldResult.setK(kFoldProp.getK());
			kFoldResult.setTotalCorrect(totalcorrect);
			kFoldResult.setTotalIncorrect(totalmissed);
			kFoldResult.setWeightCounts(weightCounts);
			
			// The total weights need to be divided by different values depending on whether a kernel was used or not.
			// If a kernel was not used, then each feature was used in every fold
			// If a kernel was used, each data point was used in every fold except one
			if (comboProperty[0].getLearnProperty().getKernel() == Constants.KERNEL_DIRECT) totalWeights.mult(1.0 / (double)kFoldProp.getK());
			else 																			totalWeights.mult(1.0 / ((double)kFoldProp.getK() - 1));
			
			output.print(ch, OutputController.TOTAL_BASIS_USED, "Average basis functions used: "+kFoldResult.getAvgBasisFunc());
			output.print(ch, OutputController.TOTAL_TIME_ELAPSED, "Total time elapsed: "+kFoldResult.getTotalTime());
			output.print(ch, OutputController.TOTAL_CORRECT, "Total correct: "+kFoldResult.getTotalCorrect());
			output.print(ch, OutputController.TOTAL_INCORRECT, "Total incorrect: "+kFoldResult.getTotalIncorrect());
			output.print(ch, OutputController.TOTAL_PERCENTAGE, "Total percentage correct: "+(double)kFoldResult.getTotalCorrect()*100/((double)kFoldResult.getTotalCorrect()+(double)kFoldResult.getTotalIncorrect())+"%");
			//output.printManual(ch, kFoldResult.getMisclassString());
			output.print(ch, OutputController.AVERAGE_WEIGHTS, "Average Weights: \n"+kFoldResult.getAverageWeights().toString());
			output.print(ch, OutputController.TOTAL_WEIGHTCOUNT, kFoldResult.getWeightCountString());
			output.print(ch, OutputController.AVERAGE_WEIGHTCOUNT, kFoldResult.getAvgWeightCountString());
			String str = "Features (one-based) to be excluded " + kFoldResult.getExcludedBases(0.5); 
			output.printManual(ch, str);

		}
		
		// Print the confusion matrix
		// This has a lot of nasty hacked formatting stuff.
		if (testOption != Constants.TEST_NONE && testOption != Constants.TEST_UNLABELED) {
			String classes[] = new String[numclasses];
			int maxlen = 3;
			for (int i = 0 ; i < numclasses; i++) {
				classes[i] = (String) classMap.getOriginalClassification(new Double(i+1));
				maxlen = Math.max(maxlen, classes[i].length());
			}
			maxlen += 2;
			
			String s;
			
			s = "\nConfusion matrix: \n";
			String out = "";
			while (out.length() < maxlen) out += " ";
			s += " " + out + " ";
			for (int i = 0; i < confu[0].length; i++) {
				out = classes[i]; 
				while (out.length() < maxlen) out += " ";
				s += "| "+out+" ";
			}
			s += "\n";
			out = "";
			while (out.length() < maxlen) out += "-";
			s += " " + out + " ";
			for (int i = 0; i < confu[0].length; i++) {
				out = "";
				while (out.length() < maxlen) out += "-";
				s += "| "+out+" ";
			}
			s += "\n";
			for (int i = 0; i < confu.length; i++) {
				out = classes[i]; 
				while (out.length() < maxlen) out += " ";
				s += " "+out+" ";
				for (int k = 0; k < confu[i].length; k++) {
					out = Integer.toString(confu[i][k]);
					while (out.length() < maxlen) out += " ";
					s += "| "+out+" ";
				}
				s += "\n";
			}
			output.print(OutputController.SUMMARY, OutputController.CONFUSION_MATRIX, s);
		}
		
		// Calculate and output the ROC results, only if it is a binary classifier
		if (binary && testOption != Constants.TEST_NONE && testOption != Constants.TEST_UNLABELED) {
			ClassifierThresholdFinder ctf = new ClassifierThresholdFinder();
			ROCResults roc = ctf.generateROC(finalProbs, finalClasses, 2);
			output.print(OutputController.SUMMARY, OutputController.ROC_CURVE, roc.printResults());
		}

		output.printManual(OutputController.BOTH, "\nEnd of learning task.");
		
		if (output.checkError()) System.out.println("Error encountered while writing output.");
	}

	/**Fills the properties with the values read from the options set and list 
	 * @param optionsSet a set of the options, used for quick in/out calculation
	 * @param options a list of the options, used for position checking
	 * @param testoption the test option constant, emulating an int& so the value can be set
	 * @param columnChoice the column choice option, emulating an int& so the value can be returned
	 * @return the file name of the desired output file, null if none given
	 * @throws ArgumentFormatException if any arguments are not valid
	 */
	public static String parseArgs(Set optionsSet, ArrayList options, ModifyProperty modifyprop[], LearnProperty learnprop, 
	        TestProperty testprop, KFoldProperty kfoldprop, OutputController output, int[] testoption, int[] columnChoice)
			throws ArgumentFormatException {
		
		if (options.size() <= 1)
			options.add("-version");
		
		if (optionsSet.contains("-help"))
			throw new ArgumentFormatException("HELP");

		if (optionsSet.contains("-behavior")) {
			int index = options.indexOf("-behavior");
			String behavior = (String) options.get(index + 1);
			if (behavior.equals("kfoldtest"))
				testoption[0] = (Constants.TEST_KFOLD);
			else if (behavior.equals("loocv"))
				testoption[0] = (Constants.TEST_LOOCV);
			else if (behavior.equals("testontesting"))
				testoption[0] = (Constants.TEST_TESTING_DATA);
			else if (behavior.equals("testontraining"))
				testoption[0] = (Constants.TEST_TRAINING_DATA);
			else if (behavior.equals("none"))
				testoption[0] = (Constants.TEST_NONE);
			else if (behavior.equals("unlabeled"))
				testoption[0] = (Constants.TEST_UNLABELED);
			else
				throw new ArgumentFormatException("Undefined behavior");
		}

		if (optionsSet.contains("-kfold")) {
			int index = options.indexOf("-kfold");
			try {
				kfoldprop.setK(Integer
						.parseInt((String) options.get(index + 1)));
			} catch (NumberFormatException e) {
				throw new ArgumentFormatException("Bad kfold parameter");
			}
		}

		if (optionsSet.contains("-kfoldbehavior")) {
			int index = options.indexOf("-kfoldbehavior");
			String next = (String) options.get(index + 1);
			int behavior = 0;
			if (next.equals("random"))
				behavior = Constants.KFOLD_PURE_RANDOM;
			if (next.equals("ratiorandom"))
				behavior = Constants.KFOLD_RATIO_RANDOM;
			if (next.equals("inplace"))
				behavior = Constants.KFOLD_NON_RANDOM;
			if (behavior != 0)
				kfoldprop.setBehavior(behavior);
			else
				throw new ArgumentFormatException("Bad kfoldbehavior parameter");
		}

		if (optionsSet.contains("-training")) {
			learnprop.setFileName(parseFileName(options, options.indexOf("-training") + 1));
		}

		if (optionsSet.contains("-testing")) {
			testprop.setTestingFileName(parseFileName(options, options.indexOf("-testing") + 1));
		}

		if (optionsSet.contains("-weights")) {
			testprop.setWeightFileName(parseFileName(options, options.indexOf("-weights") + 1));
		}
		
		if (optionsSet.contains("-unlabeled")) {
			testprop.setTestingFileName(parseFileName(options, options.indexOf("-unlabeled") + 1));
		}
		
		String outFile = null;
		if (optionsSet.contains("-saveto")) {
			outFile = parseFileName(options, options.indexOf("-saveto") + 1);
		}

		if (optionsSet.contains("-lambda")) {
			int index = options.indexOf("-lambda");
			try {
				learnprop.setLambda(Double.parseDouble((String) options
						.get(index + 1)));
			} catch (NumberFormatException e) {
				throw new ArgumentFormatException("Bad lambda");
			}
		}

		if (optionsSet.contains("-convergencetol")) {
			int index = options.indexOf("-convergencetol");
			try {
				learnprop.setConvTol(Double.parseDouble((String) options
						.get(index + 1)));
			} catch (NumberFormatException e) {
				throw new ArgumentFormatException("Bad convergence tolerance");
			}
		}

		if (optionsSet.contains("-killthreshold")) {
			int index = options.indexOf("-killthreshold");
			try {
				learnprop.setKillThreshold(Double.parseDouble((String) options
						.get(index + 1)));
			} catch (NumberFormatException e) {
				throw new ArgumentFormatException("Bad kill threshold");
			}
		}

		if (optionsSet.contains("-kernelparam")) {
			int index = options.indexOf("-kernelparam");
			try {
				learnprop.setKernelPar(Double.valueOf((String) options
						.get(index + 1)));
			} catch (NumberFormatException e) {
				throw new ArgumentFormatException("Bad kernel parameter");
			}
		}

		if (optionsSet.contains("-maxiterations")) {
			int index = options.indexOf("-maxiterations");
			try {
				learnprop.setMaxIters(Integer.parseInt((String) options
						.get(index + 1)));
			} catch (NumberFormatException e) {
				throw new ArgumentFormatException("Bad max iterations");
			}
		}

		if (optionsSet.contains("-kernel")) {
			int index = options.indexOf("-kernel");
			String next = (String) options.get(index + 1);
			int kernel = 0;
			if (next.equals("rbf"))
				kernel = Constants.KERNEL_RBF;
			if (next.equals("direct"))
				kernel = Constants.KERNEL_DIRECT;
			if (next.equals("linear"))
				kernel = Constants.KERNEL_LINEAR;
			if (next.equals("cosine"))
				kernel = Constants.KERNEL_COSINE;
			if (next.equals("polynomial"))
				kernel = Constants.KERNEL_POLYNOMIAL;
			if (kernel != 0)
				learnprop.setKernel(kernel);
			else
				throw new ArgumentFormatException("Bad kernel");
		}

		if (optionsSet.contains("-prior")) {
			int index = options.indexOf("-prior");
			String next = (String) options.get(index + 1);
			int prior = 0;
			if (next.equals("laplacian"))
				prior = Constants.PRIOR_LAPLACIAN;
			if (next.equals("gaussian"))
				prior = Constants.PRIOR_GAUSSIAN;
			if (next.equals("none"))
				prior = Constants.PRIOR_NONE;
			if (prior != 0)
				learnprop.setPrior(prior);
			else
				throw new ArgumentFormatException("Bad prior");
		}

		if (optionsSet.contains("-algorithm")) {
			int index = options.indexOf("-algorithm");
			String next = (String) options.get(index + 1);
			int algorithm = 0;
			if (next.equals("noncomponent"))
				algorithm = Constants.ALGORITHM_NON_COMPONENT;
			if (next.equals("component"))
				algorithm = Constants.ALGORITHM_COMPONENT;
			if (algorithm != 0)
				learnprop.setMethod(algorithm);
			else
				throw new ArgumentFormatException("Bad algorithm");
		}

		if (optionsSet.contains("-overrelaxed")) {
			int index = options.indexOf("-overrelaxed");
			String next = (String) options.get(index + 1);
			int overrelaxed = 0;
			if (next.equals("true"))
				overrelaxed = 1;
			if (next.equals("false"))
				overrelaxed = 2;

			if (overrelaxed != 0)
				learnprop.setUseOverrelaxation(overrelaxed == 1);
			else
				throw new ArgumentFormatException("Bad overrelaxed parameter");
		}
		
		if (optionsSet.contains("-cacheexp")) {
			int index = options.indexOf("-cacheexp");
			String next = (String) options.get(index+1);
			int cacheexp = 0;
			if (next.equals("true")) 
				cacheexp = 1;
			if (next.equals("false"))
				cacheexp = 2;
			if (cacheexp != 0)
			    learnprop.setCacheExp(cacheexp == 1);
			else
			    throw new ArgumentFormatException("Bad Cache Exponential parameter");
		}
		if (optionsSet.contains("-cachekernel")) {
			int index = options.indexOf("-cachekernel");
			String next = (String) options.get(index+1);
			int cachekernel = 0;
			if (next.equals("true")) 
				cachekernel = 1;
			if (next.equals("false"))
				cachekernel = 2;
			if (cachekernel != 0)
			    learnprop.setCacheKer(cachekernel == 1);
			else
			    throw new ArgumentFormatException("Bad Cache Kernel parameter");
		}

		if (optionsSet.contains("-bias")) {
			int index = options.indexOf("-bias");
			String next = (String) options.get(index + 1);
			if (next.equals("false")) learnprop.setBias(0);
			else { 
				try {
					learnprop.setBias(Double.parseDouble(next));
				}
				catch (NumberFormatException e) {
					throw new ArgumentFormatException("Bad bias parameter");
				}
			}
		}

		if (optionsSet.contains("-deleterowstrain")) {
			int index = options.indexOf("-deleterowstrain") + 1;

			while (!((String) options.get(index)).startsWith("-")) {
				String curr = (String) options.get(index);
				try {
					modifyprop[0].getDeletedRows().add(
							new Integer(Integer.parseInt(curr)-1));
					index++;
				} catch (NumberFormatException e) {
					throw new ArgumentFormatException("Bad row to delete");
				}
			}
		}

		if (optionsSet.contains("-deletecolumnstrain")) {
			int index = options.indexOf("-deletecolumnstrain") + 1;

			while (!((String) options.get(index)).startsWith("-")) {
				String curr = (String) options.get(index);
				try {
					modifyprop[0].getDeletedColumns().add(
							new Integer(Integer.parseInt(curr)-1));
					index++;
				} catch (NumberFormatException e) {
					throw new ArgumentFormatException("Bad column to delete");
				}
			}
		}
		
		if (optionsSet.contains("-deleterowstest")) {
			int index = options.indexOf("-deleterowstest") + 1;

			while (!((String) options.get(index)).startsWith("-")) {
				String curr = (String) options.get(index);
				try {
					modifyprop[1].getDeletedRows().add(
							new Integer(Integer.parseInt(curr)-1));
					index++;
				} catch (NumberFormatException e) {
					throw new ArgumentFormatException("Bad row to delete");
				}
			}
		}

		if (optionsSet.contains("-deletecolumnstest")) {
			int index = options.indexOf("-deletecolumnstest") + 1;

			while (!((String) options.get(index)).startsWith("-")) {
				String curr = (String) options.get(index);
				try {
					modifyprop[1].getDeletedColumns().add(
							new Integer(Integer.parseInt(curr)-1));
					index++;
				} catch (NumberFormatException e) {
					throw new ArgumentFormatException("Bad column to delete");
				}
			}
		}
		
		if (optionsSet.contains("-deleterowsunlabeled")) {
			int index = options.indexOf("-deleterowsunlabeled") + 1;

			while (!((String) options.get(index)).startsWith("-")) {
				String curr = (String) options.get(index);
				try {
					modifyprop[2].getDeletedRows().add(
							new Integer(Integer.parseInt(curr)-1));
					index++;
				} catch (NumberFormatException e) {
					throw new ArgumentFormatException("Bad row to delete");
				}
			}
		}

		if (optionsSet.contains("-deletecolumnsunlabeled")) {
			int index = options.indexOf("-deletecolumnsunlabeled") + 1;

			while (!((String) options.get(index)).startsWith("-")) {
				String curr = (String) options.get(index);
				try {
					modifyprop[2].getDeletedColumns().add(
							new Integer(Integer.parseInt(curr)-1));
					index++;
				} catch (NumberFormatException e) {
					throw new ArgumentFormatException("Bad column to delete");
				}
			}
		}

		if (optionsSet.contains("-normalize")) {
			int index = options.indexOf("-normalize");
			String next = (String) options.get(index + 1);
			int normalize = 0;
			if (next.equals("true"))
				normalize = 1;
			if (next.equals("false"))
				normalize = 2;

			if (normalize != 0)
				learnprop.setNormalize(normalize == 1);
			else
				throw new ArgumentFormatException("Bad normalize parameter");
		}
	
		if (optionsSet.contains("-choosecolumn")) {
			int index = options.indexOf("-choosecolumn") + 1;
			String next = (String) options.get(index);
			try {
				int col = Integer.parseInt(next) - 1;
				if (col >= 0) columnChoice[0] = (col);
				else columnChoice[0] = (Constants.READER_GUESS);
			}
			catch (NumberFormatException e) {
				if (next.equals("first"))
					columnChoice[0] = (Constants.READER_FIRST_COLUMN);
				else if (next.equals("last"))
					columnChoice[0] = (Constants.READER_LAST_COLUMN);
				else if (next.equals("guess"))
					columnChoice[0] = (Constants.READER_GUESS);
				else
					throw new ArgumentFormatException("Invalid argument for -choosecolumn");
			}
		}

		if (optionsSet.contains("-displayweights")) {
			int index = options.indexOf("-displayweights") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.WEIGHTS, true);
			else if (next.equals("false"))
				output.setControl(OutputController.WEIGHTS, false);
			else
				throw new ArgumentFormatException("Invalid argument for -displayweights");
		}

		if (optionsSet.contains("-displayiterations")) {
			int index = options.indexOf("-displayiterations") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.ITERATIONS, true);
			else if (next.equals("false"))
				output.setControl(OutputController.ITERATIONS, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayiterations");
		}

		if (optionsSet.contains("-displaybasisused")) {
			int index = options.indexOf("-displaybasisused") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.BASIS_USED, true);
			else if (next.equals("false"))
				output.setControl(OutputController.BASIS_USED, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displaybasisused");
		}

		if (optionsSet.contains("-displaytime")) {
			int index = options.indexOf("-displaytime") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.TIME_ELAPSED, true);
			else if (next.equals("false"))
				output.setControl(OutputController.TIME_ELAPSED, false);
			else
				throw new ArgumentFormatException("Invalid argument for -displaytime");
		}

		if (optionsSet.contains("-displayprobabilities")) {
			int index = options.indexOf("-displayprobabilities") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.PROBABILITIES, true);
			else if (next.equals("false"))
				output.setControl(OutputController.PROBABILITIES, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayprobabilities");
		}

		if (optionsSet.contains("-displayclassifications")) {
			int index = options.indexOf("-displayclassifications") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.CLASSIFICATIONS, true);
			else if (next.equals("false"))
				output.setControl(OutputController.CLASSIFICATIONS, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayclassifications");
		}

		if (optionsSet.contains("-displaynumwrong")) {
			int index = options.indexOf("-displaynumwrong") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.CORRECT, true);
			else if (next.equals("false"))
				output.setControl(OutputController.CORRECT, false);
			else
				throw new ArgumentFormatException("Invalid argument for -displaynumwrong");
		}

		if (optionsSet.contains("-displaynumcorrect")) {
			int index = options.indexOf("-displaynumcorrect") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.INCORRECT, true);
			else if (next.equals("false"))
				output.setControl(OutputController.INCORRECT, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displaynumcorrect");
		}

		if (optionsSet.contains("-displaypercentage")) {
			int index = options.indexOf("-displaypercentage") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.PERCENTAGE, true);
			else if (next.equals("false"))
				output.setControl(OutputController.PERCENTAGE, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displaypercentage");
		}
		
		if (optionsSet.contains("-displaymisclassifications")) {
			int index = options.indexOf("-displaymisclassifications") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.MISCLASSIFICATIONS, true);
			else if (next.equals("false"))
				output.setControl(OutputController.MISCLASSIFICATIONS, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displaymisclassifications");
		}
		
		if (optionsSet.contains("-displaytotalcorrect")) {
			int index = options.indexOf("-displaytotalcorrect") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.TOTAL_CORRECT, true);
			else if (next.equals("false"))
				output.setControl(OutputController.TOTAL_CORRECT, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displaytotalcorrect");
		}
		
		if (optionsSet.contains("-displaytotalwrong")) {
			int index = options.indexOf("-displaytotalwrong") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.TOTAL_INCORRECT, true);
			else if (next.equals("false"))
				output.setControl(OutputController.TOTAL_INCORRECT, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displaytotalwrong");
		}
		
		if (optionsSet.contains("-displaytotalpercentage")) {
			int index = options.indexOf("-displaytotalpercentage") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.TOTAL_PERCENTAGE, true);
			else if (next.equals("false"))
				output.setControl(OutputController.TOTAL_PERCENTAGE, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displaytotalpercentage");
		}
		
		if (optionsSet.contains("-displayaverageweights")) {
			int index = options.indexOf("-displayaverageweights") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.AVERAGE_WEIGHTS, true);
			else if (next.equals("false"))
				output.setControl(OutputController.AVERAGE_WEIGHTS, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayaverageweights");
		}
		
		if (optionsSet.contains("-displayweightcounts")) {
			int index = options.indexOf("-displayweightcounts") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.TOTAL_WEIGHTCOUNT, true);
			else if (next.equals("false"))
				output.setControl(OutputController.TOTAL_WEIGHTCOUNT, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayweightcounts");
		}
		
		if (optionsSet.contains("-displayaverageweightcounts")) {
			int index = options.indexOf("-displayaverageweightcounts") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.AVERAGE_WEIGHTCOUNT, true);
			else if (next.equals("false"))
				output.setControl(OutputController.AVERAGE_WEIGHTCOUNT, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayaverageweightcounts");
		}
		
		if (optionsSet.contains("-displayconfusionmatrix")) {
			int index = options.indexOf("-displayconfusionmatrix") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.CONFUSION_MATRIX, true);
			else if (next.equals("false"))
				output.setControl(OutputController.CONFUSION_MATRIX, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayconfusionmatrix");
		}
		
		if (optionsSet.contains("-displayroccurve")) {
			int index = options.indexOf("-displayroccurve") + 1;
			String next = (String) options.get(index);
			if (next.equals("true"))
				output.setControl(OutputController.ROC_CURVE, true);
			else if (next.equals("false"))
				output.setControl(OutputController.ROC_CURVE, false);
			else
				throw new ArgumentFormatException(
						"Invalid argument for -displayroccurve");
		}
		
		if (optionsSet.contains("-version"))
			throw new ArgumentFormatException("SMLR version 1.0.1\n"
					+ "(c) 2006 Duke University");
		return outFile;
	}
	
	/**
	 * This sets the defaults to the given parameters.
	 * @param modifyprop
	 * @param learnprop
	 * @param testprop
	 * @param kfoldprop
	 * @param dispprop
	 * @param columnChoice
	 * @return
	 */
	public static int setDefaults(ModifyProperty modifyprop[],
			LearnProperty learnprop, TestProperty testprop,
			KFoldProperty kfoldprop, OutputController output, int[] columnChoice) {

		learnprop.setNormalize(Defaults.NORMALIZE);
		learnprop.setBias(Defaults.BIAS);
		learnprop.setConvTol(Defaults.PREFS_CONVTOL);
		learnprop.setKernel(Defaults.KERNEL);
		if (Defaults.KERNEL == Constants.KERNEL_POLYNOMIAL)
			learnprop.setKernelPar(new Double(Defaults.KERNEL_POLY_PARAM));
		else if (Defaults.KERNEL == Constants.KERNEL_RBF)
			learnprop.setKernelPar(new Double(Defaults.KERNEL_RBF_PARAM));
		else
			learnprop.setKernelPar(new Double(1));
		learnprop.setKillThreshold(Defaults.PREFS_KILLTHRESH);
		learnprop.setLambda(Defaults.LAMBDA);
		learnprop.setMaxIters(Defaults.PREFS_MAXITER);
		learnprop.setMethod(Defaults.ALGORITHM);
		learnprop.setPrior(Defaults.PRIOR);
		learnprop.setUseOverrelaxation(Defaults.PREFS_OVERRELAX);
		learnprop.setFileName("No File Given");
		learnprop.setCacheExp(Defaults.PREFS_CACHEEXP);
		learnprop.setCacheKer(Defaults.PREFS_CACHEKERNEL);

		testprop.setBias(Defaults.BIAS);
		learnprop.setKernel(Defaults.KERNEL);
		if (Defaults.KERNEL == Constants.KERNEL_POLYNOMIAL)
			testprop.setKernelPar(new Double(Defaults.KERNEL_POLY_PARAM));
		else if (Defaults.KERNEL == Constants.KERNEL_RBF)
			testprop.setKernelPar(new Double(Defaults.KERNEL_RBF_PARAM));
		else
			testprop.setKernelPar(new Double(1));
		testprop.setTestingFileName("No File Given");

		kfoldprop.setBehavior(Defaults.BEHAVIOR);

		columnChoice[0] = (Defaults.CLASS_LABELS);
		
		output.setDefaults();
		
		return Defaults.BEHAVIOR;
	}
		
	/**
	 * 
	 * 
	 * @param settings lower-priority settings from settings file
	 * @param mainOptionsSet overriding settings from command line
	 * @param mainOptions overriding settings from command line
	 */
	public static void doSettings(String[] settings, HashSet mainOptionsSet, ArrayList mainOptions) throws ArgumentFormatException, HaltingException{
		int maxdig = getDigits(settings.length);
	    for (int i = 0; i < settings.length; ++i) {
			ArrayList options = new ArrayList(Arrays.asList(settings[i].split(" ")));
			HashSet optionsSet = new HashSet(options);
			options.add("-[null string]");

			LearnProperty lProp = new LearnProperty();
			TestProperty tProp = new TestProperty();
			ModifyProperty mProp[] = new ModifyProperty[3]; for (int k = 0; k < 3; k++) mProp[k] = new ModifyProperty();
			KFoldProperty kProp = new KFoldProperty();
			OutputController output = new OutputController();
			
			int[] testOption = new int[1];
			int[] columnChoice = new int[1];
			
			testOption[0] = (Core.setDefaults(mProp, lProp, tProp, kProp, output, columnChoice));

			// this gets the options out of the settings file
			String outFile = parseArgs(optionsSet, options, mProp, lProp, tProp, kProp, output, testOption, columnChoice);

			// this overrides the settings file options with the options from the command line
			String file = null;
			if(mainOptionsSet != null && mainOptions != null)
			    file = parseArgs(mainOptionsSet, mainOptions, mProp, lProp, tProp, kProp, output, testOption, columnChoice);
			
			if(file != null) outFile = file;
			String id = Integer.toString(i+1);
			while (id.length() < maxdig) id="0"+id;
			try {
				Core.process(mProp, lProp, tProp, kProp, testOption[0], output, columnChoice[0], outFile, settings[i], id);
			}
			catch (ArgumentFormatException e){
				throw new ArgumentFormatException("File Not Found");
			}
			catch (HaltingException e) {
				throw e;
			}
		}
	}
	
	private static int getDigits(int n) {
		int ret = 0;
		while (n > 0) {
			n /= 10;
			ret++;
		}
		return ret;
	}
}
