package edu.duke.cs.smlr.util;

import java.util.*;
import java.io.*;

import edu.duke.cs.smlr.factories.FileStuff;

/**
 * Consolidates all of the output information for a learning task, and is responsible for
 * printing the information
 * 
 * @author Jason Bosko
 *
 */
public class OutputController {
	
	// Which file to output to
	public static final int LOG = 1;
	public static final int SUMMARY = 2;
	public static final int BOTH = 3;
	
	// Describe the type of task being run
	public static final String LEARN_PROPERTY = "lprop";
	public static final String KFOLD_PROPERTY = "kprop";
	public static final String MODIFY_PROPERTY = "mprop";
	public static final String TEST_PROPERTY = "tprop";
	public static final String BEHAVIOR = "behave";
	
	// Describe the results of one learn
	public static final String BASIS_USED = "basis";
	public static final String TIME_ELAPSED = "time";
	public static final String WEIGHTS = "weights";
	public static final String ITERATIONS = "iters";
	
	// Describe the results of one test
	public static final String CORRECT = "correct";
	public static final String INCORRECT = "incorrect";
	public static final String PERCENTAGE = "percent";
	public static final String MISCLASSIFICATIONS = "miss";
	public static final String PROBABILITIES = "probs";
	public static final String CLASSIFICATIONS = "classes";
	
	// Describe the results of an entire cross-validation
	public static final String TOTAL_TIME_ELAPSED = "totaltime";
	public static final String TOTAL_BASIS_USED = "totalbasis";
	public static final String TOTAL_CORRECT = "totalcorrect";
	public static final String TOTAL_INCORRECT = "totalincorrect";
	public static final String TOTAL_PERCENTAGE = "totalpercent";
	public static final String TOTAL_WEIGHTCOUNT = "weightcount";
	public static final String AVERAGE_WEIGHTCOUNT = "avgcount";
	public static final String AVERAGE_WEIGHTS = "avgweight";
	public static final String CONFUSION_MATRIX = "confusion";
	public static final String ROC_CURVE = "roc";
	
	private HashMap infoMap;
	private HashMap controlMap;
	private PrintWriter pwLog;
	private PrintWriter pwSum;
	
	/**
	 * 
	 */
	public OutputController() {
		infoMap = new HashMap();
		controlMap = new HashMap();
		setDefaults();
	}
	
	/** 
	 * Sets the default controlling behavior
	 *
	 */
	public void setDefaults() {
		controlMap.put(LEARN_PROPERTY, new Boolean(true));
		controlMap.put(KFOLD_PROPERTY, new Boolean(true));
		controlMap.put(MODIFY_PROPERTY, new Boolean(true));
		controlMap.put(TEST_PROPERTY, new Boolean(true));
		controlMap.put(BEHAVIOR, new Boolean(true));
		
		controlMap.put(BASIS_USED, new Boolean(Defaults.PREFS_LEARN_BASES));
		controlMap.put(TIME_ELAPSED, new Boolean(Defaults.PREFS_LEARN_TIME));
		controlMap.put(WEIGHTS, new Boolean(Defaults.PREFS_LEARN_WEIGHTS));
		controlMap.put(ITERATIONS, new Boolean(Defaults.PREFS_LEARN_ITERS));

		controlMap.put(CORRECT, new Boolean(Defaults.PREFS_TEST_CORRECT));
		controlMap.put(INCORRECT, new Boolean(Defaults.PREFS_TEST_WRONG));
		controlMap.put(PERCENTAGE, new Boolean(Defaults.PREFS_TEST_PERCENT));
		controlMap.put(MISCLASSIFICATIONS, new Boolean(Defaults.PREFS_TEST_MISCLASS));
		controlMap.put(PROBABILITIES, new Boolean(Defaults.PREFS_TEST_PROBS));
		controlMap.put(CLASSIFICATIONS, new Boolean(Defaults.PREFS_TEST_CLASSES));
		
		controlMap.put(TOTAL_TIME_ELAPSED, new Boolean(true));
		controlMap.put(TOTAL_BASIS_USED, new Boolean(true));
		controlMap.put(TOTAL_CORRECT, new Boolean(Defaults.PREFS_CROSS_CORRECT));
		controlMap.put(TOTAL_INCORRECT, new Boolean(Defaults.PREFS_CROSS_INCORRECT));
		controlMap.put(TOTAL_PERCENTAGE, new Boolean(Defaults.PREFS_CROSS_PERCENT));
		controlMap.put(TOTAL_WEIGHTCOUNT, new Boolean(Defaults.PREFS_CROSS_WEIGHTCOUNT));
		controlMap.put(AVERAGE_WEIGHTS, new Boolean(Defaults.PREFS_CROSS_AVGWEIGHT));
		controlMap.put(AVERAGE_WEIGHTCOUNT, new Boolean(Defaults.PREFS_CROSS_NONZERO));
		controlMap.put(CONFUSION_MATRIX, new Boolean(Defaults.PREFS_CROSS_CONFUSION));
		controlMap.put(ROC_CURVE, new Boolean(Defaults.PREFS_CROSS_ROC));
	}
	
	/**
	 * Specifies the file to print log information to, <b>logFile</b>, and the file 
	 * to print summary information to, <b>summaryFile</b>. 
	 * 
	 * @param logFile The file to print log information to
	 * @param summaryFile The file to print summary information to
	 */
	public void setFiles(String logFile, String summaryFile) {
		if (logFile != null) pwLog = FileStuff.output(logFile);
		if (summaryFile != null) pwSum = FileStuff.output(summaryFile);
	}
	
	/**
	 * Specifies the file to print log information to, <b>logFile</b>, and the file 
	 * to print summary information to, <b>summaryFile</b>. 
	 * 
	 * @param logFile The file to print log information to
	 * @param summaryFile The file to print summary information to
	 */
	public void setFiles(String summaryFile) {
		pwLog = FileStuff.output(summaryFile);
		pwSum = pwLog;
	}
	
	/**
	 * Sets the control over whether information of type <b>type</b> should be printed
	 * @param type The type of information in question
	 * @param isDisplay Whether the information should be printed
	 */
	public void setControl(String type, boolean isDisplay) {
		controlMap.put(type, new Boolean(isDisplay));
	}
	
	/**
	 * Sets a piece of output for the controller. The type of information being set is 
	 * specified by <b>type</b>, and the information itself is <b>info</b>
	 * @param type The type of information being given -- use one of the static strings provided in this class
	 * @param info The information being given
	 */
	public void setOutput(String type, String info) {
		infoMap.put(type, info);
	}
	
	/**
	 * Print the information associated with <b>type</b> to the file(s) if that type 
	 * of information has been set to be printed.
	 * @param which Which file (LOG, SUMMARY, or BOTH) to print to
	 * @param type The type of information to be printed
	 */
	public void print(int which, String type) {
		if ( ((Boolean)controlMap.get(type)).booleanValue() ) { 
			if (which == LOG || which == BOTH) pwLog.println(infoMap.get(type));
			if (which == SUMMARY || which == BOTH) pwSum.println(infoMap.get(type));
		}
	}
	
	/**
	 * Prints <b>message</b> to the file(s) if <b>type</b> has been set to be printed.
	 * @param which Which file (LOG, SUMMARY, or BOTH) to print to
	 * @param type The type of information to be printed
	 * @param message The message to be printed
	 */
	public void print(int which, String type, String message) {
		if ( ((Boolean)controlMap.get(type)).booleanValue() ) { 
			if (which == LOG || which == BOTH) pwLog.println(message);
			if (which == SUMMARY || which == BOTH) pwSum.println(message);
		}
	}
	
	
	/**
	 * Automatically prints <b>message</b> to the file(s)
	 * @param which Which file (LOG, SUMMARY, or BOTH) to print to
	 * @param message The message to be printed
	 */
	public void printManual(int which, String message) {
		if (which == LOG || which == BOTH) pwLog.println(message);
		if (which == SUMMARY || which == BOTH) pwSum.println(message);
	}
	
	public void flush(int which) {
		if (which == LOG || which == BOTH) pwLog.flush(); 
		if (which == SUMMARY || which == BOTH) pwSum.flush();
	}

	public boolean checkError() {
		return pwLog.checkError() || pwSum.checkError();
	}
}
