package edu.duke.cs.smlr.util.matrix;

import java.util.AbstractList;
import java.util.ArrayList;

import edu.duke.cs.smlr.learn.interfaces.IMatrix;

/**A matrix class that supports methods for resizing and manipulating matrices.
 * This is the fourth and last abstract class used to implement the FullMatrix
 * interface.  Concrete classes should extend this class and implement the remaining
 * abstract methods, optionally overriding methods (from any of the abstract matrix
 * classes) that it can accelerate due to its particular backend implementation.
 * 
 * @author Matt Edwards
 */

public abstract class ManipMatrix extends MathMatrix implements IMatrix {
	
	public ManipMatrix() {
		super();
	}
	
	public ManipMatrix(IMatrix other) {
		super(other);
	}
	
	private ManipMatrix(int r, int c) {
	}
	
	public IMatrix chunk(int startRow, int startCol, int endRow, int endCol)
	throws ArrayIndexOutOfBoundsException {
		if(startRow < 0 || startCol < 0 || endRow > numRows() || endCol > numCols())
			throw new ArrayIndexOutOfBoundsException("Bad indices passed to chunk: {" + startRow + ", " + startCol + ", " + endRow + ", " + endCol + "} with size {" + numRows() + ", " + numCols() + "}");
		IMatrix ans = like(endRow - startRow, endCol - startCol);
		for(int i = 0; i < ans.numRows(); ++i)
			for(int j = 0; j < ans.numCols(); ++j)
				ans.set(get(i + startRow, j + startCol), i, j);
		return ans;
	}
	
	public IMatrix chunk(AbstractList rowIndices, AbstractList colIndices) throws ArrayIndexOutOfBoundsException {	    
		IMatrix ans = like(rowIndices.size(), colIndices.size());
		int rindex, cindex;
		for (int k = 0; k < rowIndices.size(); ++k) {
			rindex = ((Integer) rowIndices.get(k)).intValue();
			for (int i = 0; i < colIndices.size(); ++i) {
				cindex = ((Integer) colIndices.get(i)).intValue();
				ans.set(get(rindex, cindex), k, i);
			}
		}
		return ans;
	}
	
	public IMatrix chunkView(final AbstractList rowIndices, final AbstractList colIndices) {
		final IMatrix finalThis = this;
		return new CernMatrix() {
			private IMatrix source = finalThis;
			private Integer[] rows = (Integer []) rowIndices.toArray(new Integer[0]);
			private Integer[] cols = (Integer []) colIndices.toArray(new Integer[0]);
			public IMatrix transpose() {
				throw new UnsupportedOperationException("Can't transpose this matrix");
			}
			public IMatrix inverse() {
				throw new UnsupportedOperationException("Can't transpose this matrix");			
			}			
			public int numRows() {
				return rows.length;
			}
			public int numCols() {
				return cols.length;
			}
			public double getQuick(int r, int c) {
				return finalThis.getQuick(rows[r].intValue(), cols[c].intValue());
			}
		};
	}
	
	public IMatrix repmat(int r, int c) throws IllegalArgumentException {
		if(r < 0 || c < 0) throw new IllegalArgumentException("Bad size arguments passed to repmat");
		int myr = this.numRows();
		int myc = this.numCols();
		IMatrix ans = this.like(r * myr, c * myc);
		for (int i = 0; i < r * myr; ++i)
			for (int j = 0; j < c * myc; ++j)
				ans.set(this.get(i % myr, j % myc), i, j);                
		return ans;
	}
	
	public IMatrix diag() {
		IMatrix ans = null;
		if (numRows() > 1) {
			ans = like(numRows(), numRows());
			for (int k = 0; k < numRows(); ++k)
				ans.set(get(k, 0), k, k);
		} else if (numCols() > 1) {
			ans = like(numCols(), numCols());
			for (int k = 0; k < numCols(); ++k)
				ans.set(get(0, k), k, k);
		}
		else ans = (IMatrix) clone();
		return ans;
	}
	
	public IMatrix reshape(int r, int c) {
		if(r * c != numRows() * numCols()) throw new IllegalArgumentException("Bad size passed to reshape");
		IMatrix ans = like(r, c);
		for (int k = 0, currx = 0, curry = 0; k < numCols(); ++k)
			for (int l = 0; l < numRows(); ++l) {
				ans.set(get(l, k), curry, currx);
				curry++;
				if (curry >= r) {
					curry = 0;
					currx++;
				}
				if (currx >= c)
					currx = 0;
			}
		return ans;
	}
	
	public IMatrix expand(AbstractList indices, int maxIndex) {
		int current = 0;
		int numElem = indices.size();
		IMatrix ans = like(maxIndex, 1);
		for (int k = 0; k < maxIndex; ++k) {
			int stopIndex;
			double stopValue;
			if (current == numElem) {
				stopIndex = maxIndex;
				stopValue = 0;
			} else {
				stopIndex = ((Integer) indices.get(current)).intValue();
				stopValue = get(current, 0);
			}
			while (k < stopIndex) {
				ans.set(0, k, 0);
				k++;
			}
			if (k != maxIndex)
				ans.set(stopValue, k, 0);
			current++;
		}
		return ans;    
	}
}
