package edu.duke.cs.smlr.util.matrix;

import edu.duke.cs.smlr.learn.interfaces.IMatrix;

/**A matrix class that supports simple operations and computations that do
 * not involve another matrix.  This is the second abstract class used to
 * implement the FullMatrix interface.
 * 
 * @author Matt Edwards
 */

public abstract class SimpleMatrix extends AbstractMatrix implements IMatrix {
    
    public SimpleMatrix() {
        super();
    }
    
    public SimpleMatrix(IMatrix other) {
        super(other);
    }
    
    private SimpleMatrix(int r, int c) {
    }
    
    /**Returns the maximum absolute valued row sum
     * @return the maximum absolute valued row sum
     */
    public int maxAbsRowSum() {
        int maxSum = 0;
        int tempSum;
        for(int i = 0; i < numRows(); ++i) {
            tempSum = 0;
            for(int j = 0; j < numCols(); ++j)
                tempSum += getQuick(i, j);
            tempSum = Math.abs(tempSum);
            if(tempSum > maxSum) maxSum = tempSum;
        }        
        return maxSum;        
    }
    
    /**Returns a column vector of row sums
     * @return a column vector of row sums
     */
    public IMatrix rowSum() {
        IMatrix ans = like(numRows(), 1);
        for(int i = 0; i < numRows(); ++i)
            for(int j = 0; j < numCols(); ++j)
                ans.setQuick(getQuick(i, j) + ans.getQuick(i, 0), i, 0);        
        
        return ans;
    }
    
    /**Returns a row vector of column sums
     * @return a row vector of column sums
     */
    public IMatrix columnSum() {
        IMatrix ans = like(1, numCols());
        for(int i = 0; i < numRows(); ++i)
            for(int j = 0; j < numCols(); ++j)
                ans.setQuick(getQuick(i, j) + ans.getQuick(0, j), 0, j);        
        
        return ans;
    }    
    
    /**Returns the sum of every element in the matrix
     * @return the sum of every element in the matrix
     */
    public double sum() {
        double sum = 0;
        for(int i = 0; i < numRows(); ++i)
            for(int j = 0; j < numCols(); ++j)
                sum += getQuick(i, j);
        return sum;
    }
    
	/**Returns a row vector of the averages of each column
	 * @return a row vector of the averages of each column
	 */
	public IMatrix columnMean() {
		IMatrix ans = like(1, numCols());
		for (int i = 0; i < numRows(); ++i)
			for (int j = 0; j < numCols(); ++j) 
				ans.setQuick(getQuick(i, j) + ans.getQuick(0, j), 0, j);
		
		for (int i = 0; i < numCols(); ++i) {
			ans.setQuick(ans.getQuick(0, i) / ((double) numRows()), 0, i);
		}
		return ans;
	}

	/**Returns a row vector of the standard deviations of each column
	 * @return a row vector of the standard deviations of each column
	 */
	public IMatrix columnStd() {
		IMatrix avg = columnMean();
		IMatrix ans = like(1, numCols());
		for (int i = 0; i < numRows(); ++i)
			for (int j = 0; j < numCols(); ++j) {
				double temp = getQuick(i, j) - avg.getQuick(0, j);
				temp *= temp;
				ans.setQuick(ans.getQuick(0, j) + temp, 0, j);
			}
		for (int i = 0; i < numCols(); i++) {
			ans.setQuick(ans.getQuick(0,i) / ((double) numRows()), 0, i);
		}
		for (int i = 0; i < numCols(); ++i) {
			ans.setQuick(Math.sqrt(ans.getQuick(0, i)), 0, i);
		}
		return ans;
	}
	
	/**Returns a column vector of the row index of the maximum absolute value in each row
	 * <br>
	 * In the case of a tie, the lowest index (first occurrence) is used
	 * @return a column vector of row indices
	 */
	public IMatrix maxIndices() {
		IMatrix ans = like(numRows(), 1);
		double max, maxIndex;
		for (int i = 0; i < numRows(); ++i) {
			max = 0;
			maxIndex = -1;
			for (int j = 0; j < numCols(); ++j)
				if (Math.abs(getQuick(i, j)) > max) {
					max = Math.abs(getQuick(i, j));
					maxIndex = j;
				}
			ans.setQuick(maxIndex, i, 0);
		}
		return ans;
	}	
	
	/**Returns a column vector of the maximum absolute value in each row
	 * @return a column vector of maximum absolute values
	 */
	public IMatrix maxEachRow() {
		IMatrix ans = like(numRows(), 1);
		double max;
		for (int i = 0; i < numRows(); ++i) {
			max = 0;
			for (int j = 0; j < numCols(); ++j)
				if (Math.abs(getQuick(i, j)) > max)
					max = Math.abs(getQuick(i, j));
			ans.setQuick(max, i, 0);
		}
		return ans;
	}
	
	/**Returns the sum of absolute values of elements in the matrix
	 * @return the sum of absolute values
	 */
	public double sumAbs() {
	    double sum = 0;
	    for(int i = 0; i < numRows(); ++i)
	        for(int j = 0; j < numCols(); ++j)
	            sum += Math.abs(getQuick(i, j));
	    return sum;
	}
	
	/**Multiplies all elements by a scalar
	 * @param d the scalar to multiply by
	 */
	public IMatrix mult(double d) {
	    for(int i = 0; i < numRows(); ++i)
	        for(int j = 0; j < numCols(); ++j)
	            setQuick(d * getQuick(i, j), i, j);
	    return this;
	}
	
	/**Adds to all elements
	 * @param d the scalar to add
	 */
	public IMatrix add(double d) {
	    for(int i = 0; i < numRows(); ++i)
	        for(int j = 0; j < numCols(); ++j)
	            setQuick(d + getQuick(i, j), i, j);
	    return this;
	}
	
	/**Raises each element to a power
	 * @param d the power to raise each element to
	 */
	public IMatrix termPow(double d) {
	    for(int i = 0; i < numRows(); ++i)
	        for(int j = 0; j < numCols(); ++j)
	            setQuick(Math.pow(getQuick(i, j), d), i, j);
	    return this;
	}
	
	/**Exponentiates each element
	 * 
	 */
	public IMatrix exp() {
	    for(int i = 0; i < numRows(); ++i)
	        for(int j = 0; j < numCols(); ++j)
	            setQuick(Math.exp(getQuick(i, j)), i, j);
	    return this;
	}
	
	/**Replaces each element with its natural logarithm
	 * 
	 */
	public IMatrix log() {
	    for(int i = 0; i < numRows(); ++i)
	        for(int j = 0; j < numCols(); ++j)
	            setQuick(Math.log(getQuick(i, j)), i, j);
	    return this;
	}
	
	/**Computes the Euclidean two-norm if this matrix is a row or column vector
	 * <br>
	 * In general, this will calculate the square root of the sum of the squares of each element in the
	 * matrix
	 * @return the square root of the sum of the squares of each element
	 */
	public double norm() {
	    double sum = 0, temp;
	    for(int i = 0; i < numRows(); ++i)
	        for(int j = 0; j < numCols(); ++j) {
	            temp = getQuick(i, j);
	            sum += temp * temp;
	        }
	    return sum;	            
	} 
}
